/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */

#include <stdio.h>
#include <bcm/bcm.h>
#include <string.h>
#undef strcpy
#undef strcmp
#undef memset
#if defined(HAVE_CONFIG_H)
  #include <config.h>
#else
  #include <bcm/bcmconfig.h>
#endif

#if defined(BCM_FOR_MC)
  // with memcheck
  #if !defined(MEMCHECKDEBUG)
    #define realmalloc (void*)__real_malloc
    #define realfree   __real_free
    #define realrealloc __real_realloc
    #define strcpy __real_strcpy
    #define strcmp __real_strcmp
    #define sprintf __real_sprintf
    #define memcpy __real_memcpy
    #define memset __real_memset
  #else
    #define realmalloc (void*)malloc
    #define realfree   free
    #define realrealloc realloc
  #endif
#elif defined(HAVE___LIBC_MALLOC)
  // linux
  extern void* __libc_malloc(size_t);
  extern void* __libc_realloc(void*, size_t);
  extern void __libc_free(void*);
  #define realmalloc (void*)__libc_malloc
  #define realrealloc (void*)__libc_realloc
  #define realfree __libc_free
#elif defined(HAVE__MALLOC)
  // win32 with cygwin
  #define realmalloc (void*)_malloc
  #define realfree _free
#elif defined(WIN32) && ! defined(__CYGWIN__)
  // native win32
  #define realmalloc realmallocptr
  #define realfree realfreeptr
  #define realrealloc realreallocptr
  void * realmallocFirst(size_t);
  void * realreallocFirst(void*, size_t);
  void realfreeFirst(void *);
  static void * realmallocSecond(size_t);
  static void * realreallocSecond(void*, size_t);
  static void realfreeSecond(void *);

  BCMMAIN_API extern void * (*realmallocptr) (size_t);
  BCMMAIN_API extern void (*realfreeptr) (void *);
  BCMMAIN_API extern void * (*realreallocptr) (void *, size_t);
#elif defined(sun)
  // no memory - hooks available
  #define realmalloc (void*)malloc
  #define realrealloc (void*)realloc
  #define realfree free
#else
  #error "unknown architecture in bcm.h !!!"
#endif

BCMMAIN_API extern void * (*mallocptr) (size_t, char *, int);
BCMMAIN_API extern void * (*reallocptr) (void *, size_t, char *, int);
BCMMAIN_API extern void (*freeptr) (void *, char *, int);
BCMMAIN_API extern void (*logptr) (char *);

// int checkMemoryBlock(void *m, char *f, int l, char *func);
// void *bCheckedMalloc(size_t x, char *file, int line);
static void bCheckedFree(void *, char *file, int line);
void defaultLog(char *msg);
static void *bCheckedRealloc(void *old, size_t x, char *file, int line);

void setLogFunction(logFuncPtr f) {
  logptr=f;
  }
void defaultLog(char *msg) {
  #if defined(BCM_FOR_MC)
  	__real_fprintf(BCMOUT, msg);
  #else
  	fprintf(BCMOUT, msg);
  #endif
  }
static void logMessage(char *msg) {
  logptr(msg);
  }
static void logMessageLine(char *msg, char* nfile, int nline) {
  char tmp[1024];
  sprintf(tmp, "%s, in file '%s', in line '%d'\n", msg, nfile, nline);
  logMessage(tmp);
  }

void bcmInit(int onoff) {
  char tmp[1024];
  if (onoff) {
    sprintf(tmp, "bcm: bcmInit(): enable checked memory functions\n");
    logMessage(tmp);
    
    mallocptr=bCheckedMalloc;
    reallocptr=bCheckedRealloc;
    freeptr=bCheckedFree;
    }
  else {
    mallocptr=uncheckedMalloc;
    freeptr=uncheckedFree;
    reallocptr=uncheckedRealloc;

    sprintf(tmp, "bcm: bcmInit(): disabled checked memory functions\n");
    logMessage(tmp);
    }
  }

void *BCMMalloc(size_t s, char *nfile, int nline) {
  return mallocptr(s, nfile, nline);
  }
void BCMFree(void *m, char *nfile, int nline) {
  freeptr(m, nfile, nline);
  }
void *BCMRealloc(void *m, size_t size, char * file, int line) {
  return reallocptr(m, size, file, line);
  }

#if ! defined(BCM_FOR_MC) && ! defined(sun)
void *malloc(size_t s) {
  return mallocptr(s, BCMUNKNOWN, 0);
  }
void *calloc(size_t size, size_t s) {
  return BCMCalloc(size, s, BCMUNKNOWN, 0);
  }
void free(void *s) {
  freeptr(s, BCMUNKNOWN, 0);
  }
void *realloc(void *m, size_t size) {
  return reallocptr(m, size, BCMUNKNOWN, 0);
  // return BCMRealloc(m, size, BCMUNKNOWN, 0);
  }
#endif

void *uncheckedMalloc(size_t s, char *nfile, int nline) {
  return realmalloc(s);
  }
void *uncheckedRealloc(void *old, size_t s, char *nfile, int nline) {
  return (void*)realrealloc(old, s);
  }
void uncheckedFree(void* s2, char *nfile, int nline) {
  realfree(s2);
  }

void *bCheckedMalloc(size_t s, char *nfile, int nline) {
  void	*m;
  bcmalloc *alloc=NULL;
  
  if (s<=0) {
    logMessageLine("malloc with size <= 0", nfile, nline);
    return NULL;
    }
  // 
  // allocate more memory than needed, add struct bcmalloc before and
  // some dummy bytes after the allocated block:
  //
  m=(void*)realmalloc(s+sizeof(bcmalloc)+sizeof(BCMDUMMY));
  if (m==NULL) {
    logMessageLine("malloc returns NULL", nfile, nline);
    return m;
    }
  alloc=(bcmalloc*)m;

  // Dummy vorne: 
  strcpy(alloc->dummy, BCMDUMMY);
  // Dummy hinten: 
  strcpy((char*)UsrMemFromMem(m)+s, BCMDUMMY);
  alloc->file=nfile;
  alloc->line=nline;
  alloc->size=s;
  strcpy(alloc->magic, BCMMAGIC);
  return UsrMemFromMem(m);
  }

void *BCMCalloc(size_t num, size_t s, char *file, int line) {
  int size;
  void *m;

  size=num*s;

  m=BCMMalloc(size, file, line);
  memset(m, 0, size);
  return m;
  }

static void *bCheckedRealloc(void *old, size_t size, char *file, int line) {
  void *neu, *m;
  bcmalloc *alloc=NULL;

  if (old==NULL) {
    // --> malloc
    // logMessageLine("realloc with NULL pointer called", file, line);
    return BCMMalloc(size, file, line);
    }

  if (size==0) {
    // == free()
    BCMFree(old, file, line);
    return NULL;
    }

  m=MemFromUsrMem(old);
  alloc=(bcmalloc*)m;

  neu=BCMMalloc(size, file, line);
  memcpy(neu, old, alloc->size>size ? size : alloc->size);
  BCMFree(old, file, line);
  return neu;
  }

int checkMemoryBlock(void *m, char* dfile, int dline, char *funcname) {
  int      ok=1;
  bcmalloc *alloc=(bcmalloc*)m;
  char     tmp[1024];

  // check magic number:
  if (strcmp(BCMMAGIC, alloc->magic)!=0) {
    sprintf(tmp, "%s: unknown memory", funcname);
    logMessageLine(tmp, dfile, dline);
    return 0;
    }
  // dummy vorne testen: 
  if (strcmp(BCMDUMMY, alloc->dummy)!=0) {
    sprintf(tmp, "%s: write before memory block", funcname);
    logMessageLine(tmp, dfile, dline);
    sprintf(tmp, "%s: memory was allocated ", funcname);
    logMessageLine(tmp, alloc->file, alloc->line);
    ok=0;
    }
  // dummy hinten testen: 
  if (strcmp(BCMDUMMY, (char*)UsrMemFromMem(m)+alloc->size)!=0) {
    sprintf(tmp, "%s: write after memory block", funcname);
    logMessageLine(tmp, dfile, dline);
    sprintf(tmp, "%s: memory was allocated ", funcname);
    logMessageLine(tmp, alloc->file, alloc->line);
    ok=0;
    }
  return ok;
  }

void randomSet(void *m, void* key, int len) {
  while (len) {
    *(char*)m=(char)key++;
    m++;
    len--;
    }
}

static void bCheckedFree(void *usrmem, char *dfile, int dline) {
  void    *m;

  if (usrmem==NULL) {
    logMessageLine("free(NULL)", dfile, dline);
    return;
    }

  m=MemFromUsrMem(usrmem);
  if (checkMemoryBlock(m, dfile, dline, "free")) {
    // this is a block from us:
    // Speicher vor dem Freigeben mit Muster Fuellen, so dass ein 
    // Wiederverwenden nach dem Freigeben zu einem (hoffentlich) Fehler fuehrt: 
    // memset(m, FILL_CHAR, ((bcmalloc*)m)->size+sizeof(bcmalloc)+sizeof(BCMDUMMY)); 
    randomSet(m, m, ((bcmalloc*)m)->size+sizeof(bcmalloc)+sizeof(BCMDUMMY));
    realfree(m);
    }
  else {
    // this block is not from us:
    realfree(usrmem);
    }
  return;
  }

#if defined(WIN32) && ! defined(__CYGWIN__)
#include <windows.h>
static HANDLE heap=NULL;
static void InitHeap() {
  if (heap==NULL)
    // heap=HeapCreate(0, 5*1024*1024, 0);
    heap=GetProcessHeap();
  if (heap==NULL)
    logMessage("bcm: can't create heap !");
  }

void * realmallocFirst(size_t s) {
  InitHeap();
  realmallocptr=realmallocSecond;
  return HeapAlloc(heap, 0, s);
  }
void * realreallocFirst(void *old, size_t s) {
  InitHeap();
  realreallocptr=realreallocSecond;
  return HeapReAlloc(heap, 0, old, s);
  }
void realfreeFirst(void *m) {
  InitHeap();
  realfreeptr=realfreeSecond;
  HeapFree(heap, 0, m);
  }

static void * realmallocSecond(size_t s) {
  return HeapAlloc(heap, 0, s);
  }
static void realfreeSecond(void *m) {
  HeapFree(heap, 0, m);
  }
static void * realreallocSecond(void *old, size_t s) {
  return HeapReAlloc(heap, 0, old, s);
  }
#endif





