/*
 * MemCheck, a memory-checker for C/C++.
 * 
 * This file is released under the terms of the GNU General Public License
 * as they appear in http://www.gnu.org/licenses/gpl.txt.
 * 
 * Copyright 2001, 2002 erwin.rieger@ibrieger.de
 */

#include <memcheck.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <limits.h>
#include <time.h>
#include <ctype.h>
#include <signal.h>
#include <dlfcn.h>
#include <dirent.h>
#include <sys/stat.h>
#include <time.h>
#include <locale.h>
#include <netdb.h>
#include <assert.h>

#if defined(__CYGWIN__)
  #include <windows.h>
  #include <winnt.h>
  #include <winrename.h>
#endif

#if defined(AIXV3)
extern char * sys_errlist[];
#endif

#define CheckRights 1

#define DebCheckStr 1 #rm 
#define DebDl 1 #rm 
#define DebSyms 1
#define DebMakeHeap 1
#define DebLoadLibs 1

 // #define DebCheckExec 1
 // #define DebMergeRightsFull 1
 // #define DebSetRightsFull 1
 // #define DebMergeRights 1
 // #define DebCheckAlloc
 // #define DebListAppend 1
 // #define DebCheckAddr 1
 // #define DebSetBm 1
 // #define DebSetRight 1

#define MCFileBlockSize 10
#define MCLineBlockSize 10
#define MCFreeBlockSize 100
#define MCListBlockSize 100
#define MCSymtabBlockSize 100
#define MCFreeage 15           /* seconds, ueber env. var setzen ! */
#define MAXBUFFER 1024
/* we don't know the len of functions, so we test if CHECKEXECLEN bytes are readable */
#define CHECKEXECLEN 4
// stack shrink before we reset stack rights
#define MINSTACKSHRINK 256
                              
// static 
mcglob *mcglobalP=NULL;

#define checkOneByte(memright, right) { \
    switch (right) { \
      case MCUN: \
        if ((memright)==MCUN) { res=1; break; } \
        break; \
      case MCRW: \
        if ((memright)==MCRW) { res=1; break; } \
        break; \
      case MCRO: \
        if (((memright)==MCRO) || ((memright)==MCRW)) { res=1; break; }; \
        break; \
      case MCWO: \
        if ((memright)==MCWO) { res=2; break; } \
        if (! ((memright)==MCRW)) break; \
        res=1; \
        break; \
      default: \
        mcabort("internal error in check_addr"); \
      } }

// some protos:
// static int checkOneByte(uchar memright, uchar right);
static void check_one_free(mcfree_list *f, int);
static int registerDynamicLib(void* handle, const char*);
MCINLINE static void checkStackWrite(const void *ptr, size_t len, ulong bp);
MCINLINE static void checkStack(ulong bp);
static void printmem(unsigned char *mm, int size);
static void list_append(mclist *, void* newelement);
static void printMemoryError(mcsegment *seg, const void *ptr, size_t len, uchar right);
// static int checknull(const void *p, char* func, char* file, int line);

static uchar mcmemright(ulong ptr, mcsegment *segment);
static int comp_files(const void *, const void *);
static int comp_lines(const void *, const void *);
static int comp_free(const void *, const void *);
static void add_file(int, char *, int);
static void add_line(mcfile *f, int line, int);
MCINLINE static mcline *  find_mem(char *, int);
static mcfile * find_file(char *);
static mcline *  find_line(mcfile *, int);
#if ! defined MEMCHECKDEBUG
MCINLINE static void add_free(bcmalloc *, int, char*, int);
#else
	static void add_free(bcmalloc *, int, char*, int);
#endif
static void checkRightA4(mcsegment *seg, const void *ptr, size_t len, uchar right);
static void checkRightUnaligned(mcsegment *seg, const void *ptr, size_t len, uchar unaligned, uchar right);
static void guessSource(const ulong ptr);

#define LIST_SORT(xxx_list, index) \
  qsort((xxx_list)->list+((xxx_list)->elsize*index), (xxx_list)->count-index, (xxx_list)->elsize, (xxx_list)->compfunc);

#define SORT_DEL_LIST(xxx_list, index, howmuch) { \
  LIST_SORT(xxx_list, index) \
  (xxx_list)->count-=howmuch; \
  }

#define RESIZE_ARR(arr, size, count) { \
  if (*arr==NULL) *arr=__real_malloc(size*count); \
  else *arr=__real_realloc(*arr, size*count); }

#define BLOCKNUMBER(ptr) ((ulong)ptr/MCHeapBlock)
#define INSTACK(ptr) ((ulong)ptr>=mcglobalP->stacklow)
#define FINDHEAPSEGMENT(segvar, ptrvar) { \
  if ((segvar=&mcglobalP->heap.segments[BLOCKNUMBER(ptrvar)])->bm==NULL) \
    segvar=NULL; }

MCINLINE static void initSegment(mcsegment *seg, uchar rights) {
#ifdef DebMakeHeap
  __real_fprintf(MCOUT, "initSegment: creating new heap seg %lx .. %lx\n", seg->addr, seg->end);
#endif
  seg->bm=__real_malloc(MCBMSIZE);
  __real_memset(seg->bm, rights, MCBMSIZE);
  }

mcglob * getMCGlobalPtr() {
  return mcglobalP;
  }
void setMCGlobalPtr(mcglob *ptr) {
  mcglobalP=ptr;
  __real_fprintf(MCOUT, "MemCheck: Setglob called %lx\n", (ulong)mcglobalP);
  }
#ifdef __CYGWIN__
struct _reent *mc_get_impure_ptr() {
  return mcglobalP->impure_ptr;
  }
#endif

void MemCheckClose(int status) {
  static int closed=0;

  if (closed++) {
    __real_fprintf(MCOUT, "\nMemCheckClose(), warning: multiple close\n");
    return;
  }

  if (mcglobalP->verbose) {
    __real_fprintf(MCOUT, "\nMemCheckClose() !\n");
    __real_fprintf(MCOUT, "Checking left allocs (%d) ...\n", mcglobalP->allocs.count);
    }
  check_allocs();
  if (mcglobalP->verbose) {
    __real_fprintf(MCOUT, "Checking delayed free's (%d) ...\n", mcglobalP->freelistcount);
    }
  check_free();
  if (mcglobalP->verbose) {
    __real_fprintf(MCOUT, "NewCount: %d; DeleteCount: %d\n", mcglobalP->ncount, mcglobalP->dcount);
    /* __real_fprintf(MCOUT, "Max. Anzahl (gleichzeitig) allokierter Objekte: %d\n", maxcount); */
    printall();
    __real_fprintf(MCOUT, "internal table-usage:\n");
    __real_fprintf(MCOUT, "%.4d (%.4d) files\n", mcglobalP->fileslen, mcglobalP->filescount);
    __real_fprintf(MCOUT, "%.4d (%.4d) delayed frees \n", mcglobalP->freelistlen, mcglobalP->freelistcount);
    // __real_fprintf(MCOUT, "%.4d (%.4d) symbols\n", mcglobalP->symtablen, mcglobalP->symtabcount);
    fflush(MCOUT);
    }
  __real_exit(status);
  }

void mcabort(char *msg) {
#ifndef unix
  int *deathPointer=NULL;
#endif
  __real_fprintf(MCOUT, "MemCheck: abort after %d error%s ...\n", mcglobalP->abort_count, (mcglobalP->abort_count>1) ? "s" : " ");
  __real_fprintf(MCOUT, "MemCheck: %s\n", msg);
  fflush(MCOUT);
#ifdef unix
  raise (SIGSEGV);
#else
  *deathPointer=0;
#endif
  }

void mcprint_libs(mclist *list) {
  int i, len;
  ulong addr;
  mclib_list* l=(mclib_list*)list->list;

  for (i=0; i<list->count; i++) {
    __real_fprintf(MCOUT, "libs[%d]: `%s`, handle: %lx, refcount: %d\n", i, l[i].filename, l[i].handle, l[i].refcnt);
    addr=l[i].textstart;
    len=l[i].bssend-l[i].textstart;
    __real_fprintf(MCOUT, "libs[%d]: memory: %lx, len %d (dez)\n", i, addr, len);
    }
  }

extern char * __real_getenv(const char *name);
int get_env_int(char *name, int deflt) {
  char *s;

  s=__real_getenv(name);
  if (s!=NULL) {
	// __real_fprintf(stderr, "env %s:%s\n", name, s);
    return atoi(s);
    }
	// __real_fprintf(stderr, "env %s:defalut :%x\n", name, deflt);
  return deflt;
  }

// needed from memcheckpp.cc:
char *get_env_string(char *name, char* deflt) {
  char *s;

  s=__real_getenv(name);
  if (s!=NULL) {
    return s;
    }
  return deflt;
}

MCINLINE static int readLine(FILE * file, char * buffer, int max) {
  if(feof(file)) {
      buffer[0] = '\0';
      return 0;
    }
  else {
      char ch;
      int i = 0;

      for(;;) {
          if ((ch=getc(file))=='\r')
            // skip, dos
            continue;
	  if(ch == EOF || ch == '\n') {
      	      buffer[i] = '\0';
      	      if(i==0) return ch == '\n';
      	      else return 1;
            }
	  else
#ifdef DebReadLine
      	      if(i < max - 1)
#endif
		  buffer[i++] = ch;
#ifdef DebReadLine
      	      else
      	          __real_fprintf(MCOUT, "buffer exceeded in readLine()\n");
#endif
        }
    }
}

/* ------------------------------------------------------------------------ */
/* support functions for strings unique table */

#if 0
int parseNmLine(char * buffer);
int parseNmLine(char * buffer) {
  /* syntax of a line: ["0x"]<hexnumber><space><string><space><C++-proto> */
  char * p;
  long addr;
  mcsymtab_list st;

  if(!(p = strtok(buffer, " \t"))) return 1;
  if(p[0] == '0' && p[1] == 'x') p += 2;	/* skip '0x' (defensive) */
#ifdef __CYGWIN__
  // symbols come from dumpbinmain script, decimal
  addr=strtol(p, 0, 10);
#else
  addr=strtol(p, 0, 16);
#endif

  if(!(p = strtok(0, " \t"))) return 1;
  if(!is_valid_symbol_type(p)) return 1;
  if(!(p = strtok(0, ""))) return 1;

  if (mcglobalP->verbose>2)
    __real_fprintf(MCOUT, "MemCheck: found symbol: %s, %lx\n", p, addr);
  st.name=strdup(p);
  st.addr=addr;
  add_symtab(&st);
  return 1;
}
#endif
// get symbol ranges from app or library
//   - code starts at the first text symbol
//   - data starts at the first data symbol
//   - bss starts at the first bss symbol, ends with last symbol
// int get_symtab(const char * name, mclib_list *lib) {
int get_symtab(int libindex) {
  mclib_list *lib=&((mclib_list*)mcglobalP->libs.list)[libindex];
  FILE * pipe;
  char buffer[MAXBUFFER];
  char * p; // , *symtype;
  ulong addr;
  // long offset=0;
  // ulong textstart=-1, textend=-1, datastart=-1, dataend=-1, bssstart=-1, bssend=0;
  ulong textstart=-1, datastart=-1, bssstart=-1;
  void *sym;

#ifdef __CYGWIN__
  __real_sprintf(buffer, "%s.nmsymbols", lib->filename);
  pipe=__real_fopen(buffer, "r" );
#else
  __real_sprintf(buffer, "nm %s %s 2>/dev/null", MCNMOPTS, lib->filename);
  pipe=popen(buffer, "r");
#endif
  if(pipe==NULL) {
#ifdef __CYGWIN__
    __real_fprintf(MCOUT, "get_symtab: can't open %s\n", buffer);
#else
    __real_fprintf(MCOUT, "get_symtab: could not execute `%s'\n", buffer);
#endif
    return 0;
    }

  /* prepare symtab */
  while (readLine(pipe, buffer, MAXBUFFER)) {
    if(!(p = strtok(buffer, " \t"))) continue;
    if(p[0] == '0' && p[1] == 'x') p += 2;	/* skip '0x' (defensive) */

#ifdef __CYGWIN__
    // symbols come from dumpbinmain script, decimal
    addr=strtol(p, 0, 10);
#else
    addr=strtol(p, 0, 16);
#endif

    if(!(p = strtok(0, " \t"))) continue;

    // if (symtype[0]=='u'||symtype[0]=='U'||symtype[0]=='w'||symtype[0]=='W') {
      // if (lib==NULL) {
        // sym=findUndefined(p);
        // __real_dlerror(); // reset dlerror
        // if ((sym!=NULL) && (offset!=(ulong)sym-addr)) 
        // if (sym!=NULL) 
          // offset=(ulong)sym - addr;
          // __real_fprintf(MCOUT, "MemCheck: get_symtab: undefined: %s, %s %lx\n", name, p, (ulong)sym);
      // 
      // }
      // continue;
    // }

    // bssend=addr;
    switch(p[0]) {
      case '?': case 'a': case 'A': case 'w': case 'W': case 'V': continue;
      case 't': case 'r': continue;
      case 'U':
        fprintf(MCOUT, "MemCheck: get_symtab: unres: %s\n", p);
		break;
      case 'T': case 'R':
        // if (textstart==-1) {
          // if (lib==NULL) {
             // textstart=findTextStart() & 0xff000000;
             // mcglobalP->textstart=textstart;
             // }
          // else
             // textstart=addr; // library
          // }
        // if ((p=strtok(0, ""))!=NULL) {
          // if (lib!=NULL) {
            // sym=(void*)__real_dlsym((void*)lib->handle, p);
            // __real_dlerror(); // reset dlerror
            // if ((sym!=NULL) && (offset!=(ulong)sym-addr)) {
              // offset=(ulong)sym - addr;
              // __real_fprintf(MCOUT, "MemCheck: get_symtab: offset: %lx\n", offset);
              // }
            // }
          // }
        break;
      case 'd': case 'D':
        if (datastart==-1) {
          datastart=addr;
          // if (lib==NULL) 
             // mcglobalP->datastart=datastart;
	  }
        // if (textend==-1)
          // textend=datastart;
        break;
      case 'b': case 'B':
        if (bssstart==-1) {
          bssstart=addr;
          // if (lib==NULL) 
             // mcglobalP->bssstart=bssstart;
	  }
        // if (textend==-1)
          // textend=datastart;
        // if (dataend==-1)
          // dataend=bssstart;
        break;
      default: 
        __real_fprintf(MCOUT, "MemCheck: unknown NM - Symbol - Type: %s\n", p);
        assert(0);
        return 0;
    }
  }

  // if (lib==NULL) 
     // mcglobalP->bssend=bssend;

#ifdef __CYGWIN__
  fclose(pipe);
#else
  pclose(pipe);
#endif

  if (textstart==-1) mcabort("no symbols found");

#ifdef DebSyms
  // __real_fprintf(MCOUT, "MC: %s: textstart: %lx, datastart: %lx, bssstart: %lx, bssend %lx\n",\
    // lib->filename, textstart+offset, datastart+offset, bssstart+offset , bssend+offset);
  __real_fprintf(MCOUT, "MC: %s: textstart: %lx, datastart: %lx, bssstart: %lx\n",\
    lib->filename, textstart, datastart, bssstart);
#endif

  // if (textend==-1) textend=bssend;
  #ifdef DebSyms
    // __real_fprintf(MCOUT, "MC: inserting CS/CONST right for %s,\nrange: %lx,%lx\n",\
      // lib->filename, textstart+offset, textend+offset);
    __real_fprintf(MCOUT, "MC: inserting CS/CONST right for %s,\nrange: %lx,%lx\n",\
      lib->filename, textstart, datastart);
  #endif
  // code could span several segments, insert right carefully:
  // textend+=offset;
  for (addr=textstart; addr<datastart; addr+=MCHeapBlock) {
    if ((addr+MCHeapBlock)>datastart)
      chkr_set_right((void*)addr, datastart-addr, MCRO);
    else
      chkr_set_right((void*)addr, MCHeapBlock, MCRO);
    }

  if (datastart!=-1) {
    // if (dataend==-1) dataend=bssend; 
    //#ifdef DebSyms
      //__real_fprintf(MCOUT, "MC: inserting DS right for %s,\nrange: %lx,%lx\n",\
        //lib->filename, datastart+offset, dataend+offset);
    //#endif
    chkr_set_right((void*)datastart, bssstart-datastart, MCRW);
    if ((bssstart-datastart)>(MCHeapBlock<<1)) mcabort("data to large");
    }

  if (bssstart!=-1) {
    // #ifdef DebSyms
      // __real_fprintf(MCOUT, "MC: inserting BSS right for %s,\nrange: %lx,%lx\n",\
        // lib->filename, bssstart, bssend);
    // #endif
	assert(0);
    // chkr_set_right((void*)bssstart, bssend-bssstart, MCRW);  // bss is initialized to 0
    // if ((bssend-bssstart)>(MCHeapBlock<<1)) mcabort("data to large");
    }

assert(0);
  // if (lib!=NULL) {
    // lib->addr=textstart;
    // lib->len=bssend-textstart;
    // }
  return 1;
}

#if !defined(MEMCHECKDEBUG)
  // asm can't be used in debugged checker (fcheck-memory-usage), define function for
  // debugged checker:
  ulong getbp1() {
    /* word at *bp is saved bp of caller: */
    // asm("movl 0(%ebp), %eax");
    return (ulong)__builtin_frame_address(1);
    }
  #define getbp1() ((ulong)__builtin_frame_address(1))

// ulong findTextStart() {
  // ulong found=0xffffffff;
// 
  // ulong ret;
  // ulong bp=getbp1();
// 
  // // while (bp<mcglobalP->main_bp) 
  // // // while (1) {
    // ret=*(ulong*)(bp+sizeof(long));
    // if (ret<found) found=ret;
    // __real_fprintf(MCOUT, "findTextStart: bp %lx, ret: %lx, %lx\n", bp, ret, found);
    // bp=*(ulong*)bp;
    // if (bp==(ulong)NULL) {
      // return found;
      // }
    // }
  // }
#endif

void check_allocs() {
  bcmalloc_list *a;
  int          i;

  for (i=0; i<mcglobalP->allocs.count; i++) {
    a=&((bcmalloc_list*)mcglobalP->allocs.list)[i];
    #ifdef DebCheckAlloc
      __real_fprintf(MCOUT, "checking: %d %lx\n", i, a->addr);
    #endif
    checkMemoryBlock((void*)a->bcm_alloc, "checking left allocs", 0, "check_allocs");
    }
  __real_fprintf(MCOUT, "\n");
  }

void printall() {
  int f, l;
  mcline *lp;
  mcfile *fp;

  for (f=0; f<mcglobalP->filescount; f++) {
    fp=&(mcglobalP->files[f]);
    for (l=0;l<fp->linescount; l++) {
      lp=&(fp->lines[l]);
      if ((lp->refcount==0) && (lp->sum!=0)) {
        __real_fprintf(MCOUT, "internal error: %s: refcount: %d, sum: %d\n", fp->file, lp->refcount, lp->sum);
        __real_exit(0);
        }
      if (lp->refcount>0) {
        __real_fprintf(MCOUT, "%s:%d: refcount: %d, sum: %d\n", fp->file, lp->line, lp->refcount, lp->sum);
        }
      }
    }
  }

void check_free() {
  int i;

  for (i=0; i<mcglobalP->freelistcount; i++) {
    check_one_free(&mcglobalP->freelist[i], 1);
    }
  }

void check_one_free(mcfree_list *f, int delit) {
  void *m;
  char str[1024];

#if defined(MEMCHECKDEBUG)
  chkr_check_addr(f->m, f->size, MCUN);
#endif

  m=__real_malloc(f->size);
  // __real_memset(m, FILL_CHAR, f->size);
  randomSet(m, f->m, f->size);
  if (memcmp(m, f->m, f->size)) {
    __real_fprintf(MCOUT, "check_one_free: memory-mump:\n");
    printmem(f->m, f->size);
    __real_sprintf(str, "mem used after free, addr: %lx, len %d, malloc was at: %s/%d, free-call: %s/%d\n", (long)f->m, f->size, f->mfile, f->mline, f->file, f->line);
    check_abort(1, str);
    }
  else 
    if (delit)
      __real_free(f->m);
  __real_free(m);
  }

void printmem(uchar *mm, int size) {
  int i;

  __real_fprintf(MCOUT, "Memory-Dump (ascii):\n");
  for (i=0; i<size; i++) {
    if (isprint(*(mm+i))) __real_fprintf(MCOUT, "%c", *(mm+i));
    else __real_fprintf(MCOUT, ".");
    }
  __real_fprintf(MCOUT, "\n");

  __real_fprintf(MCOUT, "Memory-Dump (hex):\n");
  for (i=0; i<size; i++) {
    __real_fprintf(MCOUT, " %x", *(mm+i));
    }
  __real_fprintf(MCOUT, "\n");
  }

MCINLINE static void add_alloc(bcmalloc *p)  {
  bcmalloc_list new;

  if (p->size<=0)
    check_abort(1, "add_alloc with size 0 \n");

  new.bcm_alloc=p;
  list_append(&mcglobalP->allocs, &new);

  /* bcmalloc: */
  chkr_set_right(((void*)p), sizeof(bcmalloc), MCUN);

  /* mem: */
  chkr_set_right(((void*)p)+sizeof(bcmalloc), p->size, MCWO);

  /* dummy: */
  chkr_set_right(((void*)p)+sizeof(bcmalloc)+p->size, sizeof(BCMDUMMY), MCUN);
  }

MCINLINE static mclib_list* findLibByName(const char* name) {
  int i;
  mclib_list *list=mcglobalP->libs.list;

  for (i=0; i<mcglobalP->libs.count; i++) {
    if (! __real_strcmp(list[i].filename, name))
      return &list[i];
    }
  return NULL;
  }

int load_libs(const char *name) {
  FILE * pipe;
  char buffer[MAXBUFFER], *p;
  void * handle;

#ifdef __CYGWIN__
  __real_sprintf(buffer, "%s.librefs", name);
  pipe=__real_fopen(buffer, "r" );
#else
  __real_sprintf(buffer, "ldd %s 2>/dev/null|cut -f2-|cut -d' ' -f3", name);
  #ifdef DebLoadLibs
    __real_fprintf(MCOUT, "ldd command: ldd %s 2>/dev/null|cut -f2-|cut -d' ' -f3", name);
  #endif
  pipe=popen(buffer, "r");
#endif
  if(pipe==NULL) {
#ifdef __CYGWIN__
    __real_fprintf(MCOUT, "load_libs: can't open %s\n", buffer);
#else
    __real_fprintf(MCOUT, "load_libs: could not execute `%s'\n", buffer);
#endif
    return 0;
    }

  while (readLine(pipe, buffer, MAXBUFFER)) {
    // prevent recursion:
    if (findLibByName(buffer)!=NULL) {
      continue;
      }
    #ifdef DebLoadLibs
    __real_fprintf(MCOUT, "loading dependency lib: `%s`\n", buffer);
    #endif
    handle=__real_dlopen(buffer, RTLD_LAZY);
    if((p=(void*)__real_dlerror())) {
      __real_fprintf(MCOUT, "load_libs: could not load `%s', '%s'\n", buffer, p);
      return 0;
      }
    if (!registerDynamicLib(handle, buffer)) {
      __real_dlclose(handle);
      return 0;
      }
  // needed for win ?
    // if (!load_libs(buffer)) {
      // __real_fprintf(MCOUT, "load_libs: could't load dependency libs of `%s'\n", buffer);
      // return 0;
      // }
    } 

#ifdef __CYGWIN__
  fclose(pipe);
#else
  pclose(pipe);
#endif

  return 1;
  }

// we assume:
//   - code (RO) begins at handle of dlopen, it ends with the data if any
//   - data , ends with bss if any
//   - bss , ends with last symbol
int registerDynamicLib(void* handle, const char *name) {
  // void *btext, *etext, *edata;
  mclib_list   lib;

  assert(0);
  lib.handle=(ulong)handle;
  lib.refcnt=0;
  lib.filename=__real_malloc(strlen(name)+1);
  __real_strcpy(lib.filename, name);
  // get_symtab(name, &lib);
  add_lib(&mcglobalP->libs, &lib);
/*
#ifdef __CYGWIN__
    if (((btext=getDllTextSegment(handle, name, &etext))==NULL)) return 0;
    // hack, cs is handle <--> etext:
    btext=handle;
#else
    if (((btext=get_sym(handle, SymBtext, name))==NULL) && \
        ((btext=get_sym(handle, SymBtext2, name))==NULL)) return 0;
    if ((etext=get_sym(handle, SymEtext, name))==NULL) return 0;
#endif
    #ifdef DebSyms
      __real_fprintf(MCOUT, "MC: inserting CS/CONST right for %s,\nrange: %lx,%lx\n",\
        name, (ulong)btext, (ulong)etext);
    #endif
    chkr_set_right(btext, etext-btext, MCRO);

#ifdef __CYGWIN__
    // no data symbols for NT !
    // give it 64k data space:
    edata=etext+0xffff;
#else
    if ((edata=get_sym(handle, SymEBSS, name))==NULL) return 0;
#endif

    #ifdef DebSyms
      __real_fprintf(MCOUT, "MC: inserting DS right for %s,\nbegin: %lx, ... %lx...\n", name, (ulong)etext, (ulong)edata);
    #endif
    chkr_set_right(etext, edata-etext, MCRW);
*/
  return 1;
  }

void add_lib(mclist* list, mclib_list* new) {
  mclib_list *s;

  if ((s=(mclib_list *)list_find(list, (void*)new))!=NULL) {
    s->refcnt++;
    #ifdef DebDl
      __real_fprintf(MCOUT, "MemCheck: add_lib() found lib, refcount now: %d ...\n", s->refcnt);
    #endif
    return;
    }
  #ifdef DebDl
    __real_fprintf(MCOUT, "MemCheck: add_lib() appending lib ...\n");
  #endif
  new->refcnt++;
  list_append(list, new);
  }

MCINLINE static int del_lib(mclist* list, void* p) {
  mclib_list search, *res;
  search.handle=(ulong)p;

assert(0);
#if 0
  if ((res=list_find(list, (void*)&search))!=NULL) {
    if (res->refcnt<=0) {
      mcabort("internal error in del_lib, refcnt <= 0");
      }
    res->refcnt--;
    #ifdef DebDl
      __real_fprintf(MCOUT, "MemCheck: del_lib() found lib %lx, refcnt now: %d\n", res->handle, res->refcnt);
    #endif
    if (res->refcnt==0) {
      chkr_set_right((void*)res->addr, res->len, MCUN);
      // don't remove entry from list since there could be references to its memory, for
      // example in our file - list:
      //  res->handle=(ulong)-1;
      //  SORT_DEL_LIST(list, 0, 1)
      #ifdef DebDl
        mcprint_libs(list);
      #endif
      }
    return 1;
    }
 #endif
  return 0;
  }

void *list_find(mclist *list, void* search) {
  return bsearch(search, (void*)list->list, list->count,
                 list->elsize, list->compfunc);
  }

void list_append(mclist *list, void* element) {

  #ifdef DebListAppend
    __real_fprintf(MCOUT, "MemCheck: list_append(), list (a/c/l/es): (%lx/%d/%d/%d)\n", (ulong)list->list, list->count, list->len, list->elsize);
    __real_fprintf(MCOUT, "MemCheck: list_append(), element: (%lx)\n", (ulong)element);
  #endif

  if (list->count>=list->len) {
    list->len+=MCListBlockSize;
    RESIZE_ARR(&list->list, list->elsize, list->len)
    }
  __real_memcpy(list->list + (list->count*list->elsize), element, list->elsize);
  list->count+=1;
  LIST_SORT(list, 0)
  }

void *get_sym(void *lib, char *name, const char *fn) {
  void *sym;
  // char *err;

  sym=(void*)__real_dlsym(lib, name);
  // do not call dlerror, we could be called from __wrap_dlsym
  // if((err=(char*)__real_dlerror())) {
    // __real_fprintf(MCOUT, "warning '%s' while getting symbol %s from dynamic lib (%s)\n", err, name, fn);
    // return NULL;
    // }
  return sym;
  }

#ifdef __CYGWIN__
void *getDllTextSegment(void *lib, const char *fn, void **etext) {
  FILE *symf;
  char buffer[MAXBUFFER];
  void *btext=(void*)-1, *adr;
  // read objdump file, load every symbol, find lowest (btext) and
  // highest (etext) address
  __real_sprintf(buffer, "%s.objdump", fn);
  symf=__real_fopen(buffer, "r");
  if(symf==NULL) {
    __real_fprintf(MCOUT, "getDllTextSegment: can't open %s\n", buffer);
    return NULL;
    }

  *etext=NULL;
  while(readLine(symf, buffer, MAXBUFFER)) {
    // __real_fprintf(MCOUT, "read objdump line: '%s'\n", buffer);
    if ((adr=get_sym(lib, buffer, fn))==NULL) return NULL;
    // __real_fprintf(MCOUT, "adr: '%s', %lx\n", buffer, (ulong)adr);
    if (adr>*etext) *etext=adr;
    if (adr<btext) btext=adr;
    }
  __real_fprintf(MCOUT, "btext/etext of '%s': %lx, %lx\n", fn, (ulong)btext, (ulong)*etext);

  fclose(symf);
  return btext;
  }
#endif

int mc_comp_int(const void *a, const void *b) {
  if ((int*)a > (int*)b) return 1;
  if ((int*)a < (int*)b) return -1;
  return 0;
  }

MCINLINE static char *mcrightstr(char r) {
  switch(r) {
    case -1: return "XX";
    case MCUN: return "UN";
    case MCRO: return "RO";
    case MCWO: return "WO";
    case MCRW: return "RW";
    // case MCWO_TO_RW: return "WO_TO_RW";
    // case MCWO_TO_RO: return "WO_TO_RO";
    default:
      return "unknown right";
      }
  }

void add_line(mcfile *f, int line, int size) {
  mcline *l;

  if ((l=find_line(f, line))==NULL) {
    // __real_fprintf(MCOUT, "not found line %d\n", line);
    if (f->linescount==f->lineslen) {
      f->lineslen+=MCLineBlockSize;
      RESIZE_ARR((void**)&f->lines, sizeof(mcline), f->lineslen)
      }
    f->lines[f->linescount].line=line;
    f->lines[f->linescount].refcount=1;
    f->lines[f->linescount].sum=size;
    f->linescount+=1;
    qsort((void*)f->lines, f->linescount, sizeof(mcline), comp_lines);
    // for (i=0; i<f->linescount; i++)
       // __real_fprintf(MCOUT, "line %d: %d\n", i, f->lines[i].line);
    }
  else {
    // __real_fprintf(MCOUT, "found line %d\n", line);
    l->sum+=size;
    l->refcount+=1;
    }
  }

void add_free(bcmalloc *alloc, int size, char* file, int line) {
  int i, start=mcglobalP->freelistcount;
  time_t      t;

  t=__real_time(NULL);

  /* freelist ist aufsteigend sortiert */
  for (i=0; i<mcglobalP->freelistcount; i++) {
    if ((t-(mcglobalP->freelist[i].bd))>MCFreeage) {
      /* __real_fprintf(MCOUT, "free at index %d, %ld\n", i, mcglobalP->freelist[i].bd); */
      start=i;
      break;
      }
    }

  if (start<mcglobalP->freelistcount) {
    for (i=start; i<mcglobalP->freelistcount; i++) {
      check_one_free(&mcglobalP->freelist[i], 1);
      }
    qsort((void*)mcglobalP->freelist, mcglobalP->freelistcount, sizeof(mcfree_list), comp_free);
    mcglobalP->freelistcount=start;
    }

  if (mcglobalP->freelistcount==mcglobalP->freelistlen) {
    mcglobalP->freelistlen+=MCFreeBlockSize;
    RESIZE_ARR((void**)&mcglobalP->freelist, sizeof(mcfree_list), mcglobalP->freelistlen)
    }
  mcglobalP->freelist[mcglobalP->freelistcount].m=(void*)alloc;
  mcglobalP->freelist[mcglobalP->freelistcount].bd=__real_time(NULL);
  mcglobalP->freelist[mcglobalP->freelistcount].size=size;
  mcglobalP->freelist[mcglobalP->freelistcount].line=line;
  mcglobalP->freelist[mcglobalP->freelistcount].file=file;
  mcglobalP->freelist[mcglobalP->freelistcount].mfile=alloc->file;
  mcglobalP->freelist[mcglobalP->freelistcount].mline=alloc->line;
  mcglobalP->freelistcount+=1;
  qsort((void*)mcglobalP->freelist, mcglobalP->freelistcount, sizeof(mcfree_list), comp_free);

  // __real_memset((void*)alloc, FILL_CHAR, size);
  randomSet((void*)alloc, (void*)alloc, size);
  }

void add_file(int size, char *file, int line) {
  mcfile *f;

  // if ((ulong)file==0x40018c97) mcabort("xxx file");

  if ((f=find_file(file))==NULL) {
    if (mcglobalP->filescount==mcglobalP->fileslen) {
      mcglobalP->fileslen+=MCFileBlockSize;
      RESIZE_ARR((void**)&mcglobalP->files, sizeof(mcfile), mcglobalP->fileslen)
      }
    f=&(mcglobalP->files[mcglobalP->filescount]);
    f->file=file;
    f->lines=NULL;
    f->linescount=0;
    f->lineslen=0;
    __real_fprintf(MCOUT, "added file: %s\n", f->file);
    add_line(f, line, size);
    mcglobalP->filescount+=1;
    qsort((void*)mcglobalP->files, mcglobalP->filescount, sizeof(mcfile), comp_files);
    }
  else {
    add_line(f, line, size);
    }
  }

int comp_files(const void *a, const void *b) {
  return __real_strcmp(((mcfile*)a)->file, ((mcfile*)b)->file);
  }

int comp_lines(const void *a, const void *b) {
  return ((mcline*)a)->line - ((mcline*)b)->line;
  }

int comp_free(const void *a, const void *b) {
  return ((mcfree_list*)b)->bd - ((mcfree_list*)a)->bd;
  }

mcfile * find_file(char *file) {
  mcfile filesearch;
  filesearch.file=file;
  return bsearch((void*)&filesearch, (void*)mcglobalP->files, mcglobalP->filescount,
                 sizeof(mcfile), comp_files);
  }

mcline * find_mem(char *file, int line) {
  mcfile *f;
  mcline *l=NULL;

  if ((f=find_file(file))==NULL) return l;

  l=find_line(f, line);
  
  return l;
  }

mcline *  find_line(mcfile *file, int line) {
  mcline linesearch;

  linesearch.line=line;
  return bsearch((void*)&linesearch, (void*)file->lines, 
                 file->linescount, sizeof(mcline), comp_lines);
  }

#if defined(MEMCHECKDEBUG)
 static void realcpBm(const void* ptr, size_t len, void* new);
#endif
void cpBm(const void *old, size_t oldsize, void *new) {
  unsigned int  noffset, ooffset, obmoffset, nbmoffset;
  mcsegment *oldseg;
  mcsegment *newseg;
  uchar oadjust, nadjust;
  register unsigned int i;
  register uchar omaskshift;
  register uchar nmaskshift;

#if defined(MEMCHECKDEBUG)
  realcpBm(old, oldsize, new);
#endif

  oadjust=(ulong)old-MCALIGN(old);
  nadjust=(ulong)new-MCALIGN(new);

  FINDHEAPSEGMENT(oldseg, old)
  // start byte in bm:
  ooffset=((ulong)old - oldseg->addr)>>2;

  FINDHEAPSEGMENT(newseg, new)
  // start byte in bm:
  noffset=((ulong)new - newseg->addr)>>2;

  for (i=0; i<oldsize; i++) {
    obmoffset=ooffset+((i+oadjust) >> 2);
    omaskshift=((i+oadjust) & 0x3)<<1;
    nbmoffset=noffset+((i+nadjust) >> 2);
    nmaskshift=((i+nadjust) & 0x3)<<1;
    newseg->bm[nbmoffset] = (newseg->bm[nbmoffset] & (~(3 << nmaskshift))) \
    | \
    (((*(oldseg->bm+obmoffset)>>omaskshift) & 3) << nmaskshift);
    #if defined(DebSetBm)
      check_abort(mcmemright((ulong)old+i, oldseg)!=mcmemright((ulong)new+i, newseg), "mreksaldfj error");
    #endif
    }
  }

void *MCRealloc(void *old, size_t size, char *file, int line) {
  void *new;
  bcmalloc *oldalloc=(bcmalloc*)MemFromUsrMem(old);

  if (old==NULL) {
    return MCMalloc(size, file, line);
    }

  if (size==0) {
    /* == free() */
    MCFree(old, 0, file, line);
    return NULL;
    }

  /* In Liste nachschauen: */
  if (find_mem(oldalloc->file, oldalloc->line)==NULL) {
    __real_fprintf(MCOUT, "MemCheck (realloc()): Freizugebendes Objekt nicht in Liste !!!\n");
    __real_fprintf(MCOUT, "MCRealloc: malloc:  file/line/size: %s/%d/%d\n", oldalloc->file, oldalloc->line, oldalloc->size);
    __real_fprintf(MCOUT, "MCRealloc: realloc: file/line/address: %s/%d/%lx\n", file, line, (long)old);
    fflush(MCOUT);
    return NULL;
    }

  new=MCMalloc(size, file, line);
  if (size < oldalloc->size) {
    // size decreased
    cpBm(old, size, new);
    #if defined(MEMCHECKDEBUG)
      mergeRights(old, size, MCRO);
    #endif
    __real_memcpy(new, old, size);
    }
  else {
    // size increased:
    cpBm(old, oldalloc->size, new);
    #if defined(MEMCHECKDEBUG)
      mergeRights(old, oldalloc->size, MCRO);
    #endif
    __real_memcpy(new, old, oldalloc->size);
   }
  MCFree(old, oldalloc->size, file, line);
  return new;
  }
void *MCRealloc_NoInfo(void *old, size_t size) {
  return MCRealloc(old, size, BCMUNKNOWN, 0);
  }
void *MCCalloc(size_t num, size_t s, char *file, int line) {
  int size;
  void *m;

  size=num*s;

  m=MCMalloc(size, file, line);
  __real_memset(m, 0, size);
  chkr_check_addr(m, size, MCWO);
  return m;
  }
void *MCCalloc_NoInfo(size_t size, size_t s) {
  return MCCalloc(size, s, BCMUNKNOWN, 0);
  }

/*
size_t *MCFread(void *ptr, size_t size, size_t nmemb, FILE *stream, char, int) {
  int size;
  char *c;

  c=__real_fgets(s, count, f);
  if (c==NULL) return c;

  size=strlen(s);
  check_addr((void*)c, size+1, MCWO_TO_RW, "fgets res", file, line);

  return c;
  }

char *MCFgets_NoInfo(char *s, int c, FILE *f) {
  return MCFgets(s, c, f, MMCCUNKNOWN, 0);
  }
*/

char *MCFgets(char *s, int count, FILE *f, char *file, int line) {
  int size;
  char *c;

  c=__real_fgets(s, count, f);
  if (c==NULL) return c;

  size=strlen(s);
  chkr_check_addr((void*)c, size+1, MCWO);

  return c;
  }
char *MCFgets_NoInfo(char *s, int c, FILE *f) {
  return MCFgets(s, c, f, BCMUNKNOWN, 0);
  }

/*
 * CheckStack, traverse stack, perform following tests:
 *   - order of saved ebp's (esp)
 *   - if stack is word aligned
 *   - check if return addresses lie in code segments
 * This is machine dependent, wintel only (yet)
 */
static void checkStack(ulong bp) {
  ulong lastbp, ret;
  int unaligned;
  mcsegment *seg;

  // __real_fprintf(MCOUT, "MEMCHECK: check stack\n");
  while (bp<mcglobalP->main_bp) {
    // #if defined(MEMCHECKDEBUG)
      // __real_fprintf(MCOUT, "deb_chkr_checkStack: ret at %lx, main_bp: %lx\n", bp+sizeof(long), mcglobalP->main_bp);
    // #endif
    lastbp=bp;
#ifdef MEMCHECKDEBUG
    mergeRights((void*)bp, sizeof(long)*2, MCRW);
#endif
    bp=*(ulong*)bp;
    if (bp<=lastbp) {
      __real_fprintf(MCOUT, "MEMCHECK: stack error frame(n-1) %lx <= frame(n) %lx\n",
        bp, lastbp);
      mcabort("Stack error");
      }
    // check if base ptr is aligned:
    if (lastbp!=MCALIGN(lastbp)) {
      __real_fprintf(MCOUT, "stack not aligned: %lx, aligned: %lx\n", lastbp, MCALIGN(lastbp));
      mcabort("Stack error");
      }
    ret=*(ulong*)(lastbp+sizeof(long));
    if (ret<mcglobalP->main_bp) {
    	// __real_fprintf(MCOUT, "MEMCHECK: ret %lx\n", ret);
    	// check some bytes at the return address to check if it is valid:
    	// chkr_check_addr(ret, CHECKEXECLEN, MCRO);
    	FINDHEAPSEGMENT(seg, ret)
    	if ((unaligned=ret & 3)) {
      	// unaligned ptr
      	checkRightUnaligned(seg, (void*)ret, CHECKEXECLEN, unaligned, MCRO);
      	if (CHECKEXECLEN>(4-unaligned))
        	checkRightA4(seg, (void*)ret+(4-unaligned), CHECKEXECLEN-(4-unaligned), MCRO);
      	}
    	else {
      	// aligned ptr
      	checkRightA4(seg, (void*)ret, CHECKEXECLEN, MCRO);
        }
      }
    }
  }


/*
 * CheckStackWrite, traverse stack, perform following tests:
 *   - check, if write to one of the bp's or return addresses
 *
 * This is machine dependent, wintel only (yet)
 */
static void checkStackWrite(const void *ptr, size_t len, ulong bp) {
  // __real_fprintf(MCOUT, "MEMCHECK: check stack write\n");
  while (bp<mcglobalP->main_bp) {
    // __real_fprintf(MCOUT, "checkStackWrite: bp %lx, ret: %lx\n", bp, ret);

    // check, if write to bp or if write to return address:
    if (INTERSECT(bp, (sizeof(long)<<1), (ulong)ptr, len)) {
      __real_fprintf(MCOUT, "write to bp/return at %lx, write range: %lx, len %d (dez)\n", bp, (ulong)ptr, len);
      mcabort("Stack error");
      }
// #if defined(MEMCHECKDEBUG)
    // __real_fprintf(MCOUT, "DEBUGGED MEMCHECK: bp %lx stacklow: %lx, high (main_bp): %lx\n", bp, mcglobalP->stacklow, mcglobalP->main_bp);
// #endif
#ifdef MEMCHECKDEBUG
    mergeRights((void*)bp, sizeof(long), MCRW);
#endif
    bp=*(ulong*)bp;
    }
  }


/*
 * Wrapper functions:
 */
void *__wrap_realloc(void *old, size_t size) {
  return MCRealloc(old, size, "__wrap_realloc", 0);
  }
void *__wrap_calloc(size_t size, size_t s) {
  return MCCalloc(size, s, "__wrap_calloc", 0);
  }
char *__wrap_fgets(char *s, int c, FILE *f) {
  return MCFgets(s, c, f, "__wrap_fgets", 0);
  }
struct dirent* __real_readdir(DIR*);
struct dirent *__wrap_readdir(DIR *dir) {
  struct dirent *tmp;

  tmp=__real_readdir(dir);
  if (tmp==NULL) return NULL;

  chkr_set_right(tmp, sizeof(struct dirent), MCRO);
  chkr_set_right(tmp->d_name, strlen(tmp->d_name)+1, MCRO);
  return tmp;
  }

#ifndef __CYGWIN__
int __real___xstat(int, const char*, struct stat*);
int __wrap___xstat(int ver, const char *FILENAME, struct stat *BUF) {
  __real_fprintf(MCOUT, "MemCheck: stat () \n");
  MCStrlen(FILENAME, "__wrap_stat, par 1", 0);

  chkr_check_addr(BUF, sizeof(struct stat), MCWO);
  return __real___xstat(ver, FILENAME, BUF);
  }

int __wrap_stat(const char *FILENAME, struct stat *BUF) {
  return __wrap___xstat(0, FILENAME, BUF);
  }
#else
// __CYGWIN__
int __wrap_stat(const char *FILENAME, struct stat *BUF) {
  __real_fprintf(MCOUT, "MemCheck: stat () \n");
  // filename should be readable:
  MCStrlen(FILENAME, "__wrap_stat, par 1", 0);
  chkr_check_addr(BUF, sizeof(struct stat), MCWO);
  return stat(FILENAME, BUF);
  }
#endif

time_t __wrap_time(time_t *RESULT) {
  if (RESULT!=NULL)
    chkr_check_addr(RESULT, sizeof(time_t), MCWO);
  return __real_time(RESULT);
  }

FILE * __wrap_fopen (const char *path, const char *mode) {
  FILE *res;
  // check params:
  MCStrlen(path, "__wrap_fopen, par 1", 0);
  MCStrlen(mode, "__wrap_fopen, par 2", 0);
  res=__real_fopen(path, mode);
  if (res!=NULL) {
    // __real_fprintf(MCOUT, "MemCheck: fopen %lx, %d \n", (ulong)res, sizeof(FILE));
    chkr_set_right(res, sizeof(FILE), MCRO);
    }
  return res;
  }

char * __wrap_getenv (const char *name) {
  char *res;
  // check params:
  MCStrlen(name, "__wrap_getenv, par 1", 0);
  res=__real_getenv(name);
  if (res!=NULL) {
    // __real_fprintf(MCOUT, "MemCheck: __wrap_getenv %lx, %d \n", (ulong)res, sizeof(FILE));
    chkr_set_right(res, strlen(res)+1, MCRW);
    }
  return res;
  }

size_t __wrap_fread (void *DATA, size_t SIZE, size_t COUNT, FILE *STREAM) {
  int i;

  i=__real_fread(DATA, SIZE, COUNT, STREAM);
  if (i<=0) return i;

  // chkr_check_addr(DATA, SIZE*COUNT, MCWO);
  chkr_check_addr(DATA, i, MCWO);

  return i;
  }

void *MCMalloc(const size_t s, char *nfile, int nline) {
  void     *usrmem;
  char str[1024];
  ulong ret;
  ulong bp=getbp1();
  int allocsource;

  if ((s>MCHeapBlock) || (s<0)) {
    __real_sprintf(str, "MemCheck: malloc with size: %d !", s);
    check_abort(1, str);
    }
  usrmem=bCheckedMalloc(s, nfile, nline);
  add_alloc((bcmalloc*)MemFromUsrMem(usrmem));
  add_file(s, nfile, nline);
  mcglobalP->ncount++;

fprintf(stderr, "start bt at: %x\n", bp);
  while (bp<mcglobalP->main_bp) {
    ret=*(ulong*)(bp+sizeof(long));
	fprintf(stderr, "ret at: %x\n", ret);
    allocsource=guessSource2((ulong)ret);
	if (allocsource!=0) {
		assert(0);
	}
    bp=*(ulong*)bp;
    // if (bp==(ulong)NULL) {
      // break;
      // }
    }

  return(usrmem);
  }

void *MCMalloc_NoInfo(size_t s) {
  return MCMalloc(s, BCMUNKNOWN, 0);
  }
void *__wrap_malloc(size_t s) {
  return MCMalloc(s, "__wrap_malloc", 0);
  }


void MCFree_NoInfo(const void *m) {
  MCFree(m, 0, BCMUNKNOWN, 0);
  }
void __wrap_free(const void *m) {
  MCFree(m, 0, "__wrap_free", 0);
  }

void MCFree(const void *mm, size_t notused_s, char *dfile, int dline) {
  bcmalloc *alloc;
  mcline *foundline;
  bcmalloc_list *search;
  int  size;
  // char *file;
  // void    *m;

  if (!checknull(mm, "delete/free", dfile, dline)) return;

  // bcm:
  alloc=MemFromUsrMem(mm);
  check_abort(!checkMemoryBlock(alloc, dfile, dline, "MCFree"), "memory corrupted or duplicate free");

  /* In Liste nachschauen: */
  if ((foundline=find_mem(alloc->file, alloc->line))==NULL) {
    __real_fprintf(MCOUT, "MemCheck (free()): Freizugebendes Objekt nicht in Liste !!!\n");
    __real_fprintf(MCOUT, "MCFree: malloc: file/line/size: %s/%d/%d\n", alloc->file, alloc->line, alloc->size);
    __real_fprintf(MCOUT, "MCFree: free: file/line/address: %s/%d/%lx\n", dfile, dline, (long)mm);
    fflush(MCOUT);
    return;
    }

  if ((search=(bcmalloc_list*)list_find(&mcglobalP->allocs, &alloc))==NULL) {
    __real_fprintf(MCOUT, "MemCheck (free()): free auf bereits freigegebenes Objekt: !!!\n");
    __real_fprintf(MCOUT, "MCFree: malloc: file/line/size: %s/%d/%d\n", alloc->file, alloc->line, alloc->size);
    __real_fprintf(MCOUT, "MCFree: free: file/line/address: %s/%d/%lx\n", dfile, dline, (long)mm);
    check_abort(1, "mcfree");
    return;
    }
  else {
    search->bcm_alloc=(bcmalloc*)-1;
    SORT_DEL_LIST(&mcglobalP->allocs, 0, 1)
    }

  foundline->refcount--;
  foundline->sum-=alloc->size;
  if (foundline->refcount<0) {
    __real_fprintf(MCOUT, "MCFree: malloc: file/line/size: %s/%d/%d\n", alloc->file, alloc->line, alloc->size);
    __real_fprintf(MCOUT, "MCFree: free: file/line: %s/%d\n", dfile, dline);
    __real_fprintf(MCOUT, "MCFree: refcount/sum: %d/%d\n", foundline->refcount, foundline->sum);
    check_abort(1, "reference count < 0 !!!");
    }
  if (mcglobalP->verbose>2) {
    __real_fprintf(MCOUT, "MCFree: malloc: file/line/size: %s/%d/%d\n", alloc->file, alloc->line, alloc->size);
    __real_fprintf(MCOUT, "MCFree: free: file/line: %s/%d\n", dfile, dline);
    __real_fprintf(MCOUT, "MCFree: refcount/sum: %d/%d\n", foundline->refcount, foundline->sum);
    }

  /* OK, Speicher verzoegert freigeben: */
  size=alloc->size+sizeof(bcmalloc)+sizeof(BCMDUMMY);

  chkr_set_right((void*)alloc, size, MCUN);
  add_free(alloc, size, dfile, dline); 

  mcglobalP->dcount++;
  return;
  }

size_t MCStrlen(const char *s, char* f, int l) {
  int i=0;
  
  if (!checknull((void*)s, "strlen", f, l)) return 0;
  while (1) {
    chkr_check_addr(s+i, 1, MCRO);
    if (*(s+i)=='\0') return i;
    i++;
    }
  }

size_t MCStrlen_NoInfo(const char *s) {
  return MCStrlen(s, BCMUNKNOWN, 0);
  }
size_t __wrap_strlen(const char *s) {
  return MCStrlen(s, "__wrap_strlen", 0);
  }

char *MCStrcpy(char *TO, const char *FROM, char* f, int l) {
  int i=0;

  if (!checknull((void*)FROM, "strcpy, par 2", f, l)) return NULL;
  if (!checknull((void*)TO, "strcpy, par 1", f, l)) return NULL;

  i=MCStrlen(FROM, f, l);
  chkr_check_addr(TO, i+1, MCWO);
  return __real_strcpy(TO, FROM);
  }
char *MCStrcpy_NoInfo(char *TO, const char *FROM) {
  return MCStrcpy(TO, FROM, BCMUNKNOWN, 0);
  }
char *__wrap_strcpy(char *TO, const char *FROM) {
  return MCStrcpy(TO, FROM, "__wrap_strcpy", 0);
  }

char *MCStrcat(char *TO, const char *FROM, char* f, int l) {
  int i, j;

  if (!checknull((void*)FROM, "strcat, par 2", f, l)) return NULL;
  if (!checknull((void*)TO, "strcat, par 1", f, l)) return NULL;

  i=MCStrlen(FROM, f, l); /* read test */
  j=MCStrlen(TO, f, l); /* read test */
  chkr_check_addr(TO, j+i+1, MCWO);
  return __real_strcat(TO, FROM);
  }
char *MCStrcat_NoInfo(char *TO, const char *FROM) {
  return MCStrcat(TO, FROM, BCMUNKNOWN, 0);
  }
char *__wrap_strcat(char *TO, const char *FROM) {
  return MCStrcat(TO, FROM, "__wrap_strcat", 0);
  }
char *__wrap_strncat(char *TO, const char *FROM, int len) {
  int i, j;

  if (!checknull((void*)FROM, "strncat, par from", "", 0)) return NULL;
  if (!checknull((void*)TO, "strncat, dest", "", 0)) return NULL;

  for (i=0; i<len; i++) {
    chkr_check_addr((void*)FROM + i, 1, MCRO);
    if (FROM[i]=='\0') break;
    }
  j=MCStrlen(TO, "", 0); /* read test */
  chkr_check_addr(TO+j, len+1, MCWO);
  return __real_strncat(TO, FROM, len);
  }

char *MCStrncpy(char *TO, const char *FROM, size_t size, char* f, int l) {
  int i;
  char str[1000];
  
  if ((size>MCHeapBlock) || (size<0)) {
    __real_sprintf(str, "MemCheck: strncpy with size: %d !", size);
    check_abort(1, str);
    }

  if (!checknull((void*)FROM, "strncpy, par 2", f, l)) return NULL;
  if (!checknull((void*)TO, "strncpy, par 1", f, l)) return NULL;

  for (i=0; i<size; i++) {
    chkr_check_addr((void*)FROM + i, 1, MCRO);
    chkr_check_addr((void*)TO + i, 1, MCWO);
    if (*(FROM+i)=='\0') break;
    }

  return __real_strncpy(TO, FROM, size);
  }
char *MCStrncpy_NoInfo(char *TO, const char *FROM, size_t size) {
  return MCStrncpy(TO, FROM, size, BCMUNKNOWN, 0);
  }
char *__wrap_strncpy(char *TO, const char *FROM, size_t size) {
  return MCStrncpy(TO, FROM, size, "__wrap_strncpy", 0);
  }

int MCStrcmp(const char *TO, const char *FROM, char* f, int l) {
  int i=0;
  if (!checknull((void*)FROM, "strcmp, par 2", f, l)) return 0;
  if (!checknull((void*)TO, "strcmp, par 1", f, l)) return 0;

  while (1) {
    chkr_check_addr((void*)FROM+i, 1, MCRO);
    chkr_check_addr((void*)TO+i, 1, MCRO);
    if ((*(TO+i)=='\0') || (*(FROM+i)=='\0')) break;
    i++;
    }

  i=strncmp(TO, FROM, i+1);
  return i;
  }
int MCStrcmp_NoInfo(const char *TO, const char *FROM) {
  return MCStrcmp(TO, FROM, BCMUNKNOWN, 0);
  }
int __wrap_strcmp(const char *TO, const char *FROM) {
  return MCStrcmp(TO, FROM, "MCStrcmp", 0);
  }

extern char * __real_tempnam(const char *dir, const char *prefix);
char * __wrap_tempnam(const char *dir, const char *prefix) {
  char *temp, *new;
  int i;

  i=MCStrlen(dir, "__wrap_tempnam, dir", 0); /* args testen */
  i=MCStrlen(prefix, "__wrap_tempnam, prefix", 0); /* args testen */

  temp=__real_tempnam(dir, prefix);

  if (temp==NULL) return temp;

  for (i=0; *((char*)(temp+i))!='\0'; i++);
  i++;
  __real_fprintf(MCOUT, "MC: tempnam: strlen(result (%lx)): %d(dez)\n", (ulong)temp, i);

  new=MCMalloc(i, "__wrap_tempnam", 0);
  chkr_set_right(temp, i, 1);
  MCStrcpy(new, temp, "__wrap_tempnam", 0);

  __real_free(temp);

  return new;
  }

extern int __real_system(const char *cmd);
int __wrap_system(const char *cmd) {
  int i;

  i=MCStrlen(cmd, "__wrap_system, cmd", 0); /* args testen */
  return __real_system(cmd);
  }

int checknull(const void *p, char* func, char* file, int line) {
  if (p==NULL) {
    __real_fprintf(MCOUT, "MC: func %s: parameter is NULL !, File %s, Line %d\n", func, file, line);
    check_abort(1, "null");
    return 0;
    }   
  else {
    return 1;
    }   
  }

// ptr is 4 byte aligned
static void checkRightA4(mcsegment *seg, const void *ptr, size_t len, uchar right) {
  ulong bmlen;
  uchar *bmptr=&seg->bm[((ulong)ptr - seg->addr)>>2];
  int i;

  bmlen=len>>2;
  for (i=0; i<bmlen; i++) {
    switch (right) {
      case MCUN:
        if (*(bmptr+i)!=0) printMemoryError(seg, ptr+(i*4), len-(i*4), right);
        break;
      case MCRO:
        if ((*(bmptr+i) & MCROBYTE)!=MCROBYTE) printMemoryError(seg, ptr+(i*4), len-(i*4), right);
        break;
      case MCWO:
        if ((*(bmptr+i) & MCWOBYTE)==MCWOBYTE) {
          *(bmptr+i)=MCRWBYTE; // ???
          #ifdef MEMCHECKDEBUG
            mergeRights(ptr+(i<<4), 4, MCRW);
          #endif
          break;
          }
        if (! (*(bmptr+i)==MCRWBYTE)) printMemoryError(seg, ptr+(i*4), len-(i*4), right);
        break;
      case MCRW:
        if (! (*(bmptr+i)==MCRWBYTE)) printMemoryError(seg, ptr+(i*4), len-(i*4), right);
        break;
      }
    }
  // some odd bytes ?
  for (i=0; i<(len-(bmlen<<2)); i++) {
    uchar res=0;
    // res=checkOneByte(((*(bmptr+bmlen))>>(i<<1)) & 3, right);
    checkOneByte(((*(bmptr+bmlen))>>(i<<1)) & 3, right);
    if (!res) printMemoryError(seg, ptr+(bmlen<<2)+i, (len-(bmlen<<2))-i, right);
    if (res==2) {
      //change to rw ???
      *(bmptr+bmlen)=(*(bmptr+bmlen) & ~(3<<(i<<1))) | (3<<(i<<1));
      #ifdef MEMCHECKDEBUG
        mergeRights(ptr+(bmlen<<4)+i, 1, MCRW);
      #endif
      }
    }
  }

MCINLINE static ulong checkSecondBlock(mcsegment *seg, const void *ptr, size_t len, uchar right) {
  mcsegment *secondseg;
  void * endptr=(void*)ptr+len;
  ulong ourlen;

  __real_fprintf(MCOUT, "checking second block %lx, seg end: %lx, len: %d (%s)\n", (ulong)ptr, seg->end, len, mcrightstr(right));
  FINDHEAPSEGMENT(secondseg, endptr)
  if (secondseg==NULL) {
    check_abort(1, "unknown address");
    }
  ourlen=(ulong)endptr - secondseg->addr;
  checkRightA4(secondseg, (void*)secondseg->addr, ourlen, right);
  return ourlen;
  }

static void printMemoryError(mcsegment *seg, const void *ptr, size_t len, uchar right) {
  char str[1024];

  print_location(ptr, len, right, seg);
  __real_sprintf(str, "memory error, while checking with right: %s (right at %lx: %s)\n", mcrightstr(right), (ulong)ptr, mcrightstr(mcmemright((ulong)ptr, seg)));
  check_abort(1, str);
  }

// check 1, 2 or 3 bytes in the bitmap
static void checkRightUnaligned(mcsegment *seg, const void *ptr, size_t len, uchar unaligned, uchar right) {
  uint bmoffset=((ulong)ptr - seg->addr)>>2;
  uchar orig=seg->bm[bmoffset];
  int i;
  uchar shift=unaligned<<1;
  for (i=0; (i<len) && (i<4-unaligned); i++) {
    uchar res=0;
    // res=checkOneByte((orig>>shift) & 3, right);
    checkOneByte((orig>>shift) & 3, right);
    if (!res) printMemoryError(seg, ptr+i, 1, right);
    if (res==2) {
      //change to rw ???
      seg->bm[bmoffset]=seg->bm[bmoffset] | (MCRW << shift);
      #ifdef MEMCHECKDEBUG
        mergeRights(ptr+i, 1, MCRW);
      #endif
      }
    shift+=2;
    }
  }
#if 0
void old_chkr_check_addr(const void *ptr, size_t len, uchar right) {
  mcsegment *seg;    // also used to get start address of stack

  if (mcglobalP->stacklow>(ulong)&seg) {
    FINDHEAPSEGMENT(seg, (void*)&seg)
    if (seg==NULL)
      initSegment(seg, MCSTACKBYTE);
    mcglobalP->stacklow=(ulong)&seg;
    }
  //else {
    //// __real_fprintf(stderr, "reset %ld stack bytes\n", (long)(&seg) - mcglobalP->stacklow);
    //chkr_set_right((void*)mcglobalP->stacklow, (long)(&seg) - mcglobalP->stacklow, MCWO);
    //}
  chkr_check_addrNoStack(ptr, len, right);
  }
#endif
// void chkr_check_addrNoStack(const void *ptr, size_t len, uchar right) {
void chkr_check_addr(const void *ptr, size_t len, uchar right) {
  ulong bp=0;
  mcsegment    *seg;
  uchar unaligned;

  #ifdef DebCheckAddr
    __real_fprintf(MCOUT, "check_addr %lx, %ld %s\n", (ulong)ptr, (ulong)len, mcrightstr(right));
  #endif
  // len > heapblock size not tested !

  FINDHEAPSEGMENT(seg, ptr)
  if (seg==NULL) {
    // initSegment(seg, MCSTACKBYTE);
    __real_fprintf(MCOUT, "MC: no rights set for memory range: %lx+%d\n", (ulong)ptr, len);
    mcabort("unknown right");
  }
  #ifdef DebCheckAddr
    __real_fprintf(MCOUT, "check_addr: found segment : %lx...%lx\n", seg->addr, seg->end);
  #endif
  #ifdef CheckRights
    if (right>3) {
      __real_fprintf(MCOUT, "MC: unknown right: %d ... \n", right);
      mcabort("unknown right");
      }
  #endif
  if ((ulong)(ptr+len) > seg->end) {
    len-=checkSecondBlock(seg, ptr, len, right);
    }

  if (mcglobalP->checkstackmode) {
	bp=getbp1();
  	checkStack(bp);
  }
  if (INSTACK(ptr) && (right!=MCRO) && (mcglobalP->checkstackmode & 1)) {
	if (!bp) bp=getbp1();
	checkStackWrite(ptr, len, bp);
	}

  if ((unaligned=(ulong)ptr & 3)) {
    // unaligned ptr
    checkRightUnaligned(seg, ptr, len, unaligned, right);
    if (len>(4-unaligned))
      checkRightA4(seg, ptr+(4-unaligned), len-(4-unaligned), right);
    }
  else {
    // aligned ptr
    checkRightA4(seg, ptr, len, right);
    }
  }
/* 
static int checkOneByte(uchar memright, uchar right) {
    #ifdef DebCheckAddr
      __real_fprintf(MCOUT, "checking '%s' with '%s'\n", mcrightstr(memright), mcrightstr(right));
    #endif
    switch (right) {
      case MCUN:
        if (memright==MCUN) return 1;
        return 0;
      case MCRW:
        if (memright==MCRW) return 1;
        return 0;
      case MCRO:
        if ((memright==MCRO) || (memright==MCRW)) return 1;
        return 0;
      case MCWO:
        if (memright==MCWO) {
	  return 2;
	  }
        if (! (memright==MCRW)) return 0;
        return 1;
      default:
        mcabort("internal error in check_addr");
        return 0;
      }
  }
*/

void chkr_check_str (const void * ptr, uchar right) {
  #ifdef DebCheckStr
    __real_fprintf(MCOUT, "MemCheck: chkr_check_str() called, %lx, r: %s\n", (ulong)ptr, mcrightstr(right));
  #endif
  chkr_check_addr(ptr, MCStrlen(ptr, BCMUNKNOWN, 0)+1, right);
  }

// set 1, 2 or 3 bytes in the bitmap
MCINLINE static void setRightUnaligned(mcsegment *seg, const void *ptr, size_t len, uchar unaligned, uchar right) {
  uint bmoffset=((ulong)ptr - seg->addr)>>2;
  uchar *bmadr=&seg->bm[bmoffset];
  uchar l=0;

  switch (unaligned) {
    case 1:
        // set second byte in bitmap
        *bmadr=(*bmadr & 0xf3) | (right<<2);
        l++;
    case 2:
        // set third byte in bitmap
        if (l<len) *bmadr=(*bmadr & 0xcf) | (right<<4);
        l++;
    case 3:
        // set fourth byte in bitmap
        if (l<len) *bmadr=(*bmadr & 0x3f) | (right<<6);
    }
  }
// ptr is 4 byte aligned
MCINLINE static void setRightA4(mcsegment *seg, const void *ptr, size_t len, uchar right) {
  uchar *bmptr=&seg->bm[((ulong)ptr - seg->addr)>>2];
  ulong bmlen;
  int i;

  right|=(right<<2) | (right<<4) | (right<<6);
  bmlen=len>>2;
  __real_memset(bmptr, right, bmlen);

  bmptr+=bmlen;
  for (i=0; i<(len-(bmlen<<2)); i++) {
    *bmptr = (*bmptr & ~(3<<(i<<1))) | right;
    }
  }
MCINLINE static ulong setSecondBlock(const void *ptr, size_t len, uchar right) {
  mcsegment *secondseg;
  void * endptr=(void*)ptr+len;
  ulong ourlen;

  __real_fprintf(MCOUT, "set second block %lx, len: %d (%s)\n", (ulong)ptr, len, mcrightstr(right));
  if (len>MCHeapBlock) mcabort("requested memory block to large");
  FINDHEAPSEGMENT(secondseg, endptr)
  if (secondseg==NULL) {
    secondseg=&mcglobalP->heap.segments[BLOCKNUMBER(endptr)];
    initSegment(secondseg, MCSTACKBYTE);
    }
  ourlen=(ulong)endptr - secondseg->addr;
  setRightA4(secondseg, (void*)secondseg->addr, ourlen, right);
  return ourlen;
  }

void chkr_set_right(const void* ptr, size_t len, uchar right) {
  mcsegment *seg;
  uchar     unaligned;
  if (mcglobalP->stacklow>(ulong)&seg) {
    FINDHEAPSEGMENT(seg, (void*)&seg)
    if (seg==NULL) {
      seg=&mcglobalP->heap.segments[BLOCKNUMBER(ptr)];
      initSegment(seg, MCSTACKBYTE);
      }
    mcglobalP->stacklow=(ulong)&seg;
    }
  else {
    // chkr_set_right((void*)mcglobalP->stacklow, (long)(&seg) - mcglobalP->stacklow, MCWO);
    int len=(long)(&seg) - mcglobalP->stacklow;
    if (len>MINSTACKSHRINK) {
    	FINDHEAPSEGMENT(seg, (void*)&seg)
    	if (seg==NULL) {
      	seg=&mcglobalP->heap.segments[BLOCKNUMBER(&seg)];
      	initSegment(seg, MCSTACKBYTE);
      	}
    	setRightA4(seg, (void*)mcglobalP->stacklow, len, MCWO);
      }
    }

  #ifdef DebSetRight
    __real_fprintf(MCOUT, "Setr %lx l %d dez r %s\n", (ulong)ptr, len, mcrightstr(right));
  #endif

  #if defined(MEMCHECKDEBUG)
    // set right also in main checker:
    mergeRights(ptr, len, right);
  #endif

  FINDHEAPSEGMENT(seg, ptr)
  if (seg==NULL) {
    seg=&mcglobalP->heap.segments[BLOCKNUMBER(ptr)];
    initSegment(seg, MCSTACKBYTE);
    }
  #ifdef DebSetRight
    __real_fprintf(MCOUT, "set_right: found segment : %lx...%lx\n", seg->addr, seg->end);
  #endif

  if ((ulong)ptr+len > seg->end)
    len-=setSecondBlock(ptr, len, right);

  #if defined(DebSetBm) && !defined(MEMCHECKDEBUG)
    __real_fprintf(MCOUT, "bm vor setright (seg: %lx): \n", (ulong)seg);
    print_location(ptr, len, right, seg);
  #endif

  if ((unaligned=(ulong)ptr & 3)) {
    // unaligned ptr
    setRightUnaligned(seg, ptr, len, unaligned, right);
    if (len>(4-unaligned))
      setRightA4(seg, ptr+(4-unaligned), len-(4-unaligned), right);
    }
  else {
    // aligned ptr
    setRightA4(seg, ptr, len, right);
    }

  #ifdef DebSetBm
    __real_fprintf(MCOUT, "bm nach setright: \n");
    #if defined(MEMCHECKDEBUG)
      if ((right==MCRO)||(right==MCRW)) print_location(ptr, len, right, seg);
    #else
      print_location(ptr, len, right, seg);
    #endif
  #endif
  #ifdef DebSetRightsFull
  if (mcmemright((ulong)ptr, seg)!=right) {
    __real_fprintf(MCOUT, "first error 2, should: %x, is: %x, %lx, %d\n", right, mcmemright((ulong)ptr, seg), (ulong)ptr, len);
    mcabort("first error 2"); 
    }
  if (mcmemright((ulong)ptr+len-1, seg)!=right) {
    __real_fprintf(MCOUT, "last error 2, should: %x, is: %x, %lx, %d\n", right, mcmemright((ulong)ptr+len-1, seg), (ulong)ptr, len);
    mcabort("lasterror 2");
    }
  #endif
  }

// copy rights, check if destination right permits write
MCINLINE static void copy_rights_readable_writable(const void* dest, const void* src, size_t len) {
	int i;
	mcsegment *srcseg, *destseg;
  uchar sright, bcount=0, readable=0;
  uchar dbmb, dmaskshift, dright;
	unsigned int srcofs, destofs, destbmindex;

	FINDHEAPSEGMENT(srcseg, src)
	assert(srcseg!=NULL);
	FINDHEAPSEGMENT(destseg, dest)
	assert(destseg!=NULL);

 	srcofs=(ulong)src  - srcseg->addr;
 	destofs=(ulong)dest  - destseg->addr;
	assert ((srcofs+len)<MCHeapBlock);
	assert ((destofs+len)<MCHeapBlock);

	for (i=0; i<len; i++) {
		// get source right
  	// sbmb=srcseg->bm[srcofs >> 2];
  	// smaskshift=(((ulong)src+i) & 3)<<1;
		// sright=(sbmb>>smaskshift) & 3;
		sright=((srcseg->bm[srcofs >> 2])>>((((ulong)src+i) & 3)<<1)) & 3;
		// check if every 4'th source byte is readable:
		if (sright & MCRO) readable=1;

		// get destination right
  	destbmindex=destofs >> 2;
  	dbmb=destseg->bm[destbmindex];
  	dmaskshift=(((ulong)dest+i) & 3)<<1;
		dright=(dbmb>>dmaskshift) & 3;
		// check if dest is writable:
		if ((dright & MCWO)==0) printMemoryError(destseg, dest+i, 1, MCRO);

		// copy right:
		// destseg->bm[destbmindex]= (dbmb & ~(3<<dmaskshift)) | (sright<<dmaskshift);
		// destseg->bm[destbmindex]= dbmb | (sright<<dmaskshift);

		switch (sright) {
			case MCUN:
				// dest will be UN
				destseg->bm[destbmindex]= dbmb & ~(3<<dmaskshift);
				break;
			case MCWO:
				// dest will be WO
				destseg->bm[destbmindex]= (dbmb & ~(3<<dmaskshift)) | (MCWO<<dmaskshift);
				break;
			case MCRO:
			case MCRW:
				// dest will be RW
				destseg->bm[destbmindex]= (dbmb & ~(3<<dmaskshift)) | (MCRW<<dmaskshift);
				break;
			default:
        __real_fprintf(MCOUT, "internal error: unknown source right %x\n", sright);
        mcabort("internal error in copy_rights_readable_writable()");
				break;
		}

		srcofs++;
		destofs++;
		bcount++;
		if (bcount==4) {
 			if (!readable) printMemoryError(srcseg, src+i-4, 4, MCRO);
			bcount=readable=0;
		}
	}
}

// Problem copying structures with padding bytes (they are not initialized) here
// 	- check if one byte of every 4 byte group of source is readable
// 	- check if len bytes of destination are writable
//	- copy rights
void chkr_copy_bitmap (void* dest, const void* src, size_t len) {
  // __real_fprintf(MCOUT, "copy_bitmap: %lx, <-- %lx, len %d(dez)\n", (ulong)dest, (ulong)src, len);
fprintf(stderr, "chkr_copy_bitmap at: %x\n", src);
  copy_rights_readable_writable(dest, src, len);
  __real_memcpy(dest, src, len);
  }

void check_abort(int bed, char *msg) {
  if (!bed) return;

  mcglobalP->abort_count++;
  if (mcglobalP->abort_count>mcglobalP->abort_ignore) {
    mcabort(msg);
    }
  else {
    __real_fprintf(MCOUT, "MC: Error %d of %d, msg: '%s'\n", mcglobalP->abort_count, mcglobalP->abort_ignore, msg);
    }
  }

void chkr_check_exec (const void *ptr) {
  /* OK, falls Read - Recht auf addresse: */
  #ifdef DebCheckExec
    __real_fprintf(MCOUT, "MemCheck: chkr_check_exec() called, %lx\n", (ulong)ptr);
  #endif
  chkr_check_addr(ptr, CHECKEXECLEN, MCRO);
  }

#define MCLINELEN 12
#define MCADDBYTES 8
#define MCMAXPRINT 50 // muss durch 2 teilbar sein

void guessSource(const ulong ptr) {
	int i;
	mclib_list *list=mcglobalP->libs.list;
	bcmalloc_list *a;

assert(0);
#if 0
	__real_fprintf(MCOUT, "MC: pointer points\n");
	if (ptr>=mcglobalP->main_bp)
		__real_fprintf(MCOUT, "  stack above main !!!\n");
	else if (ptr>=mcglobalP->stacklow)
		__real_fprintf(MCOUT, "  in stack segment\n");
	else if (ptr>=mcglobalP->bssend) {
		__real_fprintf(MCOUT, "  in heap segment, checking allocs, libs, freelist\n");

		// allocs:
		for (i=0; i<mcglobalP->allocs.count; i++) {
			a=&((bcmalloc_list*)mcglobalP->allocs.list)[i];
			// printf("alloc prt %x, addr %x, to: %x\n", ptr, (ulong)a->bcm_alloc, (ulong)a->bcm_alloc+a->bcm_alloc->size+sizeof(bcmalloc)+sizeof(BCMDUMMY));
			if ((ptr>=(ulong)a->bcm_alloc) && (ptr<(ulong)a->bcm_alloc+(a->bcm_alloc->size)+sizeof(bcmalloc)+sizeof(BCMDUMMY))) {
				__real_fprintf(MCOUT, "  block allocated in '%s', line %d, mem %x, size %d\n", a->bcm_alloc->file, a->bcm_alloc->line, (ulong)a->bcm_alloc+sizeof(bcmalloc), a->bcm_alloc->size);
				return;
			}
		}

		// libs:
		for (i=0; i<mcglobalP->libs.count; i++) {
			// printf("lib prt %x, addr %x, to: %x\n", ptr, (ulong)list[i].addr, (ulong)list[i].addr+(ulong)list[i].len);
			if ((ptr>=(ulong)list[i].addr) && (ptr<(ulong)list[i].addr+(ulong)list[i].len)) {
				__real_fprintf(MCOUT, "  in library, not implemented\n");
				return;
			}
		}

  		// freelist 
  		for (i=0; i<mcglobalP->freelistcount; i++) {
			// printf("free prt %x, addr %x, to: %x\n", ptr, (ulong)mcglobalP->freelist[i].m, (ulong)mcglobalP->freelist[i].m+mcglobalP->freelist[i].size);
    		if ((ptr>=(ulong)mcglobalP->freelist[i].m) && (ptr<(ulong)mcglobalP->freelist[i].m+mcglobalP->freelist[i].size)) {
				__real_fprintf(MCOUT, "  in heap freelist, not implemented\n");
      			return;
      		}
    	}
	}
	else if (ptr>=mcglobalP->bssstart)
		__real_fprintf(MCOUT, "  in bss segment, symbol check not implemented\n");
	else if (ptr>=mcglobalP->datastart)
		__real_fprintf(MCOUT, "  in data segment, symbol check not implemented\n");
	else if (ptr>=mcglobalP->textstart)
		__real_fprintf(MCOUT, "  in text segment, symbol check not implemented\n");
	else
		__real_fprintf(MCOUT, "  below text segment !!!\n");
#endif
}

#include <libgen.h> // basename
int guessSource2(const ulong ptr) {
	int i;
	mclib_list *list=mcglobalP->libs.list;
	bcmalloc_list *a;
	char *bn;

assert(0);
#if 0
	__real_fprintf(MCOUT, "MC: pointer points\n");
	if (ptr>=mcglobalP->main_bp) {
		__real_fprintf(MCOUT, "  stack above main !!!\n");
		assert(0);
		}
	else if (ptr>=mcglobalP->stacklow) {
		__real_fprintf(MCOUT, "  in stack segment\n");
		assert(0);
		}
	else if (ptr>=mcglobalP->bssend) {
		__real_fprintf(MCOUT, "  in heap segment, checking allocs, libs, freelist\n");

		// libs:
		for (i=0; i<mcglobalP->libs.count; i++) {
			// printf("lib prt %x, addr %x, to: %x\n", ptr, (ulong)list[i].addr, (ulong)list[i].addr+(ulong)list[i].len);
			if ((ptr>=(ulong)list[i].addr) && (ptr<(ulong)list[i].addr+(ulong)list[i].len)) {

				// save lib ??
				if (mcglobalP->checked_libs!=NULL) {
					bn=basename(list[i].filename);
					fprintf(MCOUT, "  in library, '%s', basename: '%s'\n", list[i].filename, bn);
					if (__real_strstr(mcglobalP->checked_libs, bn)==NULL) {
						// this a unchecked lib
						fprintf(MCOUT, "  unchecked, '%s', basename: '%s'\n", list[i].filename, bn);
						return 1;
					}
					return 0; // normal checks apply
				}
				fprintf(MCOUT, "  unchecked, '%s', basename: '%s'\n", list[i].filename, bn);
				return 1; 
			}
		}

		// allocs:
		for (i=0; i<mcglobalP->allocs.count; i++) {
			a=&((bcmalloc_list*)mcglobalP->allocs.list)[i];
			// printf("alloc prt %x, addr %x, to: %x\n", ptr, (ulong)a->bcm_alloc, (ulong)a->bcm_alloc+a->bcm_alloc->size+sizeof(bcmalloc)+sizeof(BCMDUMMY));
			if ((ptr>=(ulong)a->bcm_alloc) && (ptr<(ulong)a->bcm_alloc+(a->bcm_alloc->size)+sizeof(bcmalloc)+sizeof(BCMDUMMY))) {
				__real_fprintf(MCOUT, "  block allocated in '%s', line %d, mem %x, size %d\n", a->bcm_alloc->file, a->bcm_alloc->line, (ulong)a->bcm_alloc+sizeof(bcmalloc), a->bcm_alloc->size);
				assert(0);
			}
		}

  		// freelist 
  		for (i=0; i<mcglobalP->freelistcount; i++) {
			// printf("free prt %x, addr %x, to: %x\n", ptr, (ulong)mcglobalP->freelist[i].m, (ulong)mcglobalP->freelist[i].m+mcglobalP->freelist[i].size);
    		if ((ptr>=(ulong)mcglobalP->freelist[i].m) && (ptr<(ulong)mcglobalP->freelist[i].m+mcglobalP->freelist[i].size)) {
				__real_fprintf(MCOUT, "  in heap freelist, not implemented\n");
				assert(0);
      		}
    	}
	}
	else if (ptr>=mcglobalP->bssstart) {
		// bss seg
		return 0; }
	else if (ptr>=mcglobalP->datastart) {
		// data seg
		return 0; }
	else if (ptr>=mcglobalP->textstart) {
		// program text seg
		return 0; }
	else {
		// below program text seg
		__real_fprintf(MCOUT, "  below text segment !!!\n");
		assert(0);
	}
#endif
}

void print_location(const void *ptr, size_t len, char right, mcsegment *seg) {
  int   split=0;
  uchar *c, memr=-1;
  char  ind[200], a;

assert(0);
#if 0
  // if ((ulong)ptr > 0x9000000) return;
  __real_fprintf(MCOUT, "MC: prgr. code start:	%lx\n", mcglobalP->textstart);
  __real_fprintf(MCOUT, "MC: prgr. data start:	%lx\n", mcglobalP->datastart);
  __real_fprintf(MCOUT, "MC: prgr. bss start:	%lx\n", mcglobalP->bssstart);
  __real_fprintf(MCOUT, "MC: prgr. bss end:	%lx\n", mcglobalP->bssend);
  __real_fprintf(MCOUT, "MC: stacklow:		%lx\n", mcglobalP->stacklow);
  __real_fprintf(MCOUT, "MC: stackhigh:		%lx\n", mcglobalP->main_bp);
  guessSource((ulong)ptr);
  __real_fprintf(MCOUT, "MC: memory and rights at %lx to %lx, len %d(dez):\n", (ulong)ptr, (ulong)ptr+len, len);
#endif

  if (len>MCMAXPRINT) {
    split=MCMAXPRINT/2;
    }

  for (c=(uchar*)ptr - MCADDBYTES; c < ((uchar*)ptr + MCADDBYTES + len); c++) {
    if (split) {
      if ((ulong)c >= ((ulong)ptr+split)) {
        c=c + (len - MCMAXPRINT);
        split=0;
        __real_fprintf(MCOUT, "-- skip --\n");
        }
      }
    if ((c>=(uchar*)ptr) && (c<((uchar*)ptr+len))) {
      __real_sprintf(ind, "%5d", c-(uchar*)ptr);
      }
    else {
      __real_strcpy(ind, "     ");
      }

    // __real_fprintf(MCOUT, "access %lx\n", (ulong)c);
    a='.';
    // can't acces such a low address, ld-linux stuff ?!
assert(0);
#if 0
    if ((ulong)c < mcglobalP->textstart+32) {
      __real_fprintf(MCOUT, "MC: address %lx to low in print_location()\n", (ulong)c);
      }
    else {
    	if (seg!=NULL)
      	memr=mcmemright((ulong)c, seg);
#if defined(MEMCHECKDEBUG)
      if (((void*)c>=ptr) && ((void*)c<(ptr+len))) {
        if (isprint(*c)) a=*c;
        __real_fprintf(MCOUT, "%lx %s: %2.2x, %c, %s", (ulong)c, ind, *c & 0xff, a, mcrightstr(memr));
        }
      else
        __real_fprintf(MCOUT, "%lx %s: ??, %c, %s", (ulong)c, ind, a, mcrightstr(memr));
#else
      if (isprint(*c)) a=*c;
      __real_fprintf(MCOUT, "%lx %s: %2.2x, %c, %s", (ulong)c, ind, *c & 0xff, a, mcrightstr(memr));
#endif
      if (seg!=NULL) {
        if ((ulong)c>=seg->end) __real_fprintf(MCOUT, ", after end of segment");
        }
      __real_fprintf(MCOUT, "\n");
      }
#endif
    }

  __real_fprintf(MCOUT, "\n");
  }


void *__wrap_dlopen(const char *filename, int flag) {
  void         *handle;
  // char         *err;
  mclib_list   *s;

assert(0);
  // check filename:
  MCStrlen(filename, "__wrap_dlopen, par 1", 0);

  #ifdef DebDl
    __real_fprintf(MCOUT, "MemCheck: dlopen() called, %s\n", filename);
  #endif

  // already loaded ?
  if ((s=findLibByName(filename))!=NULL) {
    __real_fprintf(MCOUT, "MC: __wrap_dlopen already loaded: '%s', handle: %lx ...\n", filename, s->handle);
    s->refcnt++;
    return (void*)s->handle;
    }
  handle=__real_dlopen(filename, flag);
  // we can't call dlerror, it resets error code and caller doesn't get
  // an error !
  // if((err=(void*)dlerror())) {
  if(handle==NULL) {
    // __real_fprintf(MCOUT, "MC: __wrap_dlopen, could not load '%s', '%s'\n", filename, err);
    __real_fprintf(MCOUT, "MC: __wrap_dlopen, could not load '%s'\n", filename);
    return handle;
    }

  if (!registerDynamicLib(handle, filename)) return NULL;

  // needed for win ?
  // if (!load_libs(filename)) {
    // __real_fprintf(MCOUT, "__wrap_dlopen: could't load dependency libs of `%s'\n", filename);
    // return NULL;
    // }

  return handle;
  }

int __wrap_dlclose(void *handle) {
  int status=0;
  #ifdef DebDl
    __real_fprintf(MCOUT, "MemCheck: dlclose() called, handle: %lx\n", (ulong)handle);
  #endif

  check_abort(!del_lib(&mcglobalP->libs, handle), 
              "dlclose: dynamic lib - handle not found !!!");
  // can't call real dlclose since there could be references to its memory, for
  // example in our file - list. Accessing this memory will result in a 'out of
  // bounds' error:
  // status=__real_dlclose(handle);
  return status;
  }

void __wrap_exit(int status) {

  // we get called from unchecked code also, for example libdl or crt0, so
  // we can't check the status:
  // check_addr(&status, 4, MCRO, "__wrap_exit", BCMUNKNOWN, 0);

  if (mcglobalP->verbose) {
    __real_fprintf(MCOUT, "MemCheck: exit(%d) called\n", status);
    }

  MemCheckClose(status); //xxx // calls exit
  // __real_exit(status);
  }

void *__wrap_memset(void *BLOCK, int C, size_t SIZE) {

  chkr_check_addr(&C, 4, MCRO);
  chkr_check_addr(&SIZE, 4, MCRO);
  chkr_check_addr(&BLOCK, 4, MCRO);
  chkr_check_addr(BLOCK, SIZE, MCWO);

  return __real_memset(BLOCK, C, SIZE);
  }


void *__wrap_memcpy(void *TO, const void *FROM, size_t SIZE) {
fprintf(stderr, "__wrap_memcpy at: %x\n", FROM);
  chkr_copy_bitmap (TO, FROM, SIZE);
  return TO;
  }

uchar mcmemright(ulong ptr, mcsegment *segment) { 
  // uchar *bmptr, maskshift;
  // bmptr=&segment->bm[(ptr - segment->addr) >> 2];
  // maskshift=((ulong)ptr & 3)*2;
  // return ((*bmptr)>>maskshift) & 3;

  return (segment->bm[(ptr - segment->addr) >> 2] >> (((ulong)ptr & 3)<<1)) & 3;
  }

/*
 * Wrapper Functions:
 */ 
extern char* __real_setlocale (int c, const char *l);
char* __wrap_setlocale (int c, const char *l) {
  char *res;

  chkr_check_addr(&c, sizeof(int), MCRO);
  if (l!=NULL) 
    MCStrlen(l, "__wrap_locale, par 2", 0);
  res=__real_setlocale(c, l);
  if (res!=NULL)
    chkr_set_right(res, strlen(res)+1, MCRO);
  return res;
  }
extern struct hostent * __real_gethostbyname(const char *h);
struct hostent * __wrap_gethostbyname(const char *h) {
  struct hostent *res;
  char   *tmp;
  int    i;

  MCStrlen(h, "__wrap_gethostbyname, par 1", 0);
  res=__real_gethostbyname(h);
  if (res!=NULL) {
    chkr_set_right(res, sizeof(struct hostent), MCRO);
    chkr_set_right(res->h_name, strlen(res->h_name)+1, MCRO);
    for (i=0; (tmp=res->h_aliases[i])!=NULL; i++) {
      chkr_set_right(res->h_aliases[i], strlen(res->h_aliases[i])+1, MCRO);
      chkr_set_right(&(res->h_aliases[i]), sizeof(char*), MCRO);
      }
    for (i=0; (tmp=res->h_addr_list[i])!=NULL; i++) {
      chkr_set_right(res->h_addr_list[i], strlen(res->h_addr_list[i])+1, MCRO);
      chkr_set_right(&(res->h_addr_list[i]), sizeof(char*), MCRO);
      }
    }
  return res;
  }

#ifdef __CYGWIN__
int __wrap_ReadFile(HANDLE h, void* buf, DWORD num, DWORD *count, 
  LPOVERLAPPED ol) {
    int res;

  chkr_check_addr(&h, 4, MCRO);
  chkr_check_addr(&buf, 4, MCRO);
  chkr_check_addr(&num, sizeof(DWORD), MCRO);
  chkr_check_addr(&count, 4, MCRO);
  chkr_check_addr(count, sizeof(DWORD), MCWO);


  res=ReadFile(h, buf, num, count, ol);
  chkr_check_addr(buf, *count, MCWO);
  return res;
  }
void *__wrap_LoadLibraryA(const char *filename) {
  return __wrap_dlopen(filename, RTLD_NOW);
  }
int __wrap_FreeLibrary(void *handle) {
  return __wrap_dlclose(handle);
  }
BOOL __wrap_CreatePipe(PHANDLE hrp, PHANDLE hwp, LPSECURITY_ATTRIBUTES attr,
  DWORD size) {

  chkr_check_addr(hrp, sizeof(HANDLE), MCWO);
  chkr_check_addr(hwp, sizeof(HANDLE), MCWO);
  // incomplete check, nested items not checked
  if (attr!=NULL)
    chkr_check_addr(attr, sizeof(LPSECURITY_ATTRIBUTES), MCRO);
  chkr_check_addr(&size, sizeof(DWORD), MCRO);
  return CreatePipe(hrp, hwp, attr, size);
  }
BOOL __wrap_CreateProcessA(LPCTSTR name, LPTSTR cmd, LPSECURITY_ATTRIBUTES pa,
  LPSECURITY_ATTRIBUTES ta, BOOL inherit, DWORD f, LPVOID env, LPCTSTR dir,
  LPSTARTUPINFO info, LPPROCESS_INFORMATION pinfo) {
    int l;

  if (name!=NULL) MCStrlen(name, "__wrap_CreateProcess, par 1", 0);
  if (cmd!=NULL) {
    l=MCStrlen(cmd, "__wrap_CreateProcess, par 2", 0);
    chkr_check_addr(cmd, l+1, MCWO);
    }
  if (pa!=NULL)
    chkr_check_addr(pa, sizeof(LPSECURITY_ATTRIBUTES), MCRO);
  if (ta!=NULL)
    chkr_check_addr(ta, sizeof(LPSECURITY_ATTRIBUTES), MCRO);
  chkr_check_addr(&inherit, sizeof(BOOL), MCRO);
  chkr_check_addr(&f, sizeof(DWORD), MCRO);
  // env is not tested !!!
  if (dir!=NULL) MCStrlen(dir, "__wrap_CreateProcess, par 8", 0);
  chkr_check_addr(info, sizeof(STARTUPINFO), MCRO);
  chkr_check_addr(pinfo, sizeof(PROCESS_INFORMATION), MCWO);

  return CreateProcess(name, cmd, pa, ta, inherit, f, env, dir, info, pinfo);
  }
DWORD __wrap_WaitForInputIdle(HANDLE h, DWORD sec) {
  chkr_check_addr(&h, sizeof(HANDLE), MCRO);
  chkr_check_addr(&sec, sizeof(DWORD), MCRO);
  return WaitForInputIdle(h, sec);
  }

int WINAPI memcheck_dll_entry(HANDLE h, DWORD reason, void *ptr) {
  return 1;
  }
asm(".section .idata$3\n" ".long 0,0,0,0, 0,0,0,0");

#endif // M$

#if defined(MEMCHECKDEBUG)
#undef cpBm
extern void cpBm(const void* ptr, size_t len, void* new);
void realcpBm(const void* ptr, size_t len, void* new) {
#ifdef DebSetRight
  __real_fprintf(MCOUT, "DebMemCheck: call real cpbm, %lx, %lx, len %ld\n", (ulong)ptr, (ulong)new, (ulong)len);
#endif
  cpBm(ptr, len, new);
  }
#else
MCINLINE static void mergeRightA4(mcsegment *seg, const void* ptr, size_t len, uchar right) {
  uint bmoffset=((ulong)ptr - seg->addr)>>2;
  uchar r, mask;
  ulong bmlen;
  int i;

  bmlen=len>>2;
  for (i=0; i<bmlen; i++) {
    switch (right) {
      case MCWO:
        if (seg->bm[bmoffset+i]==0)
          seg->bm[bmoffset+i]=MCWOBYTE;
        break;
      case MCRO:
        if (seg->bm[bmoffset+i]!=MCRWBYTE)
          seg->bm[bmoffset+i]=MCRWBYTE;
        break;
      case MCRW:
        seg->bm[bmoffset+i]=MCRWBYTE;
        break;
      }
    }
  r=right;
  for (i=bmlen<<2; i<len; i++) {
    // __real_fprintf(MCOUT, "doing last merge %d, len %d\n", i, len);
    mask=3 << (2 * (i-(bmlen<<2)));
    switch (right) {
      case MCWO:
        if ((seg->bm[bmoffset+bmlen] & mask)==0)
          seg->bm[bmoffset+bmlen]=(seg->bm[bmoffset+bmlen] & ~mask) | r;
        break;
      case MCRO:
        if (((seg->bm[bmoffset+bmlen] & mask) >> (2 * (i-(bmlen<<2))))!=MCRW)
          seg->bm[bmoffset+bmlen]=(seg->bm[bmoffset+bmlen] & ~mask) | (MCRW<< (2 * (i-(bmlen<<2))));
        break;
      case MCRW:
        seg->bm[bmoffset+bmlen]=seg->bm[bmoffset+bmlen] | r;
        break;
      }
    r=r << 2;
    // __real_fprintf(MCOUT, "done last merge %d, len %d, m %x, r, %x\n", i, len, mask, r);
    }
  }
MCINLINE static ulong mergeSecondBlock(mcsegment *seg, const void *ptr, size_t len, uchar right) {
  mcsegment *secondseg;
  void * endptr=(void*)ptr+len;
  ulong ourlen;

  __real_fprintf(MCOUT, "merge second block %lx, seg end: %lx, len: %d (%s)\n", (ulong)ptr, seg->end, len, mcrightstr(right));
  FINDHEAPSEGMENT(secondseg, endptr)
  if (secondseg==NULL) {
    secondseg=&mcglobalP->heap.segments[BLOCKNUMBER(endptr)];
    initSegment(secondseg, MCSTACKBYTE);
    }
  ourlen=(ulong)endptr - secondseg->addr;
  mergeRightA4(secondseg, (void*)secondseg->addr, ourlen, right);
  return ourlen;
  }

MCINLINE static void mergeRightUnaligned(mcsegment *seg, const void *ptr, size_t len, uchar unaligned, uchar right) {
  uint bmoffset=((ulong)ptr - seg->addr)>>2;
  uchar orig=seg->bm[bmoffset];
  int i;
  uchar shift=unaligned<<1;
  for (i=0; (i<len) && (i<4-unaligned); i++) {
    switch (right) {
      case MCWO:
        if ((orig & (3 << shift))==0) {
          seg->bm[bmoffset]=(seg->bm[bmoffset] & ~(3 << shift)) | (right<<shift);
          }
        break;
      case MCRO:
        // if (((orig & (3 << shift))>>shift)!=MCRW)
        if (((orig>>shift) & 3)!=MCRW)
          seg->bm[bmoffset]=(seg->bm[bmoffset] & ~(3 << shift)) | (MCRW<<shift);
        break;
      case MCRW:
        seg->bm[bmoffset]=seg->bm[bmoffset] | (right<<shift);
        break;
      }
    shift+=2;
    }
  }

void mergeRights(const void* ptr, size_t len, uchar right) {
  mcsegment *seg;
  uchar unaligned;
  #ifdef DebMergeRightsFull
    uchar oldfirst, oldlast;
  #endif

  #ifdef DebMergeRights
    __real_fprintf(MCOUT, "MergeRights %lx l %d dez r %s\n", (ulong)ptr, len, mcrightstr(right));
  #endif
  if (right==MCUN) return;

  FINDHEAPSEGMENT(seg, ptr)
  if (seg==NULL) {
    seg=&mcglobalP->heap.segments[BLOCKNUMBER(ptr)];
    initSegment(seg, MCSTACKBYTE);
    }
  #ifdef DebMergeRights
    __real_fprintf(MCOUT, "mergeRights: found segment : %lx...%lx\n", seg->addr, seg->end);
  #endif
  #ifdef DebMergeRightsFull
    oldfirst=mcmemright((ulong)ptr, seg);
    oldlast=mcmemright((ulong)ptr+len-1, seg);
  #endif
  if ((ulong)(ptr+len) > seg->end) {
    len-=mergeSecondBlock(seg, ptr, len, right);
    }
  if ((unaligned=(ulong)ptr & 3)) {
    // unaligned ptr
    mergeRightUnaligned(seg, ptr, len, unaligned, right);
    // __real_fprintf(MCOUT, "mergeRights, done %d unaligned, len now: %d  \n", 4-unaligned, len-(4-unaligned));
    if (len>(4-unaligned))
      mergeRightA4(seg, ptr+(4-unaligned), len-(4-unaligned), right);
    }
  else {
    // aligned ptr
    mergeRightA4(seg, ptr, len, right);
    }

  // mergeBm(&seg->bm, ((ulong)ptr)-seg->addr, len, (uchar)right);

  #if defined(DebMergeRights)
    __real_fprintf(MCOUT, "mergeRights, bm nach setright: \n");
    print_location(ptr, len, right, seg);
  #endif
  #ifdef DebMergeRightsFull
    if ((oldfirst & MCWO) && (!(mcmemright((ulong)ptr, seg) & MCWO))) {
      __real_fprintf(MCOUT, "first error, previous: %x, is: %x\n", oldfirst, mcmemright((ulong)ptr, seg));
      mcabort("first error"); 
      }
    if ((oldlast & MCWO) && (!(mcmemright((ulong)ptr+len-1, seg) & MCWO))) {
      __real_fprintf(MCOUT, "lasterror, previous: %x, is: %x\n", oldlast, mcmemright((ulong)ptr+len-1, seg));
      mcabort("lasterror"); 
      }
  #endif
  }
#endif // MEMCHECKDEBUG

